from math import sqrt, log, pi, isclose, exp, factorial
import numpy as np
from poisson_approval.utils.Util import isnan, isneginf


# noinspection NonAsciiCharacters
class Asymptotic:
    r"""An asymptotic development of the form :math:`\exp(\mu n + \nu \log n + \xi + o(1))`.

    Parameters
    ----------
    mu : Number, ``np.nan`` or ``- np.inf``
        Coefficient of the term in `n` (called "magnitude").
    nu : Number, ``np.nan`` or ``- np.inf``
        Coefficient of the term in `log n`.
    xi : Number, ``np.nan`` or ``- np.inf``
        Constant coefficient.

    Attributes
    ----------
    μ : Number, ``np.nan`` or ``- np.inf``
        Alias for :attr:`mu`.
    ν : Number, ``np.nan`` or ``- np.inf``
        Alias for :attr:`nu`.
    ξ : Number, ``np.nan`` or ``- np.inf``
        Alias for :attr:`xi`.

    Notes
    -----
    If a coefficient is ``np.nan``, it means that it is not known. In contrast, a coefficient 0 means that there is no
    corresponding term in the asymptotic development.

    If (and only if) the event is impossible, then ``mu = nu = xi = - np.inf``.

    Examples
    --------

        >>> asymptotic = Asymptotic(mu=-1 / 9, nu=-1 / 2, xi=-1 / 2 * log(8 * pi / 9))
        >>> print(asymptotic)
        exp(- 0.111111 n - 0.5 log n - 0.513473 + o(1))
        >>> asymptotic.mu
        -0.1111111111111111
        >>> asymptotic.nu
        -0.5
        >>> asymptotic.xi
        -0.5134734250965083
    """

    def __init__(self, mu, nu, xi):
        self.mu = mu
        self.nu = nu
        self.xi = xi
        self.μ = mu
        self.ν = nu
        self.ξ = xi

    def __repr__(self):
        """Repr.

        Returns
        -------
        str

        Examples
        --------
            >>> from fractions import Fraction
            >>> Asymptotic(mu=- Fraction(1, 10), nu=np.nan, xi=3)
            Asymptotic(mu=Fraction(-1, 10), nu=nan, xi=3)
        """
        return 'Asymptotic(mu=%r, nu=%r, xi=%r)' % (self.mu, self.nu, self.xi)

    def __str__(self):
        """Str.

        Returns
        -------
        str

        Examples
        --------
            >>> from fractions import Fraction
            >>> print(Asymptotic(mu=- Fraction(1, 10), nu=np.nan, xi=3))
            exp(- 0.1 n + ? log n + 3 + o(1))
            >>> print(Asymptotic(mu=-np.inf, nu=-np.inf, xi=-np.inf))
            exp(- inf)
        """
        if isneginf(self.mu) and isneginf(self.nu) and isneginf(self.xi):
            return "exp(- inf)"

        def nice(x, suffix):
            if isnan(x):
                return ' + ? ' + suffix if suffix else ' + ?'
            if x == 1 and suffix:
                return ' + ' + suffix
            if x == -1 and suffix:
                return ' - ' + suffix
            if x == 0:
                return ''
            result = ' + ' if x > 0 else ' - '
            result += "{:.6g}".format(float(abs(x)))
            result += ' ' + suffix if suffix else ''
            return result
        s = nice(self.mu, 'n') + nice(self.nu, 'log n') + nice(self.xi, '')
        s += ' + o(1)'
        if s[1] == "+":
            s = s[3:]
        else:
            s = s[1:]
        s = "exp(%s)" % s
        return s

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    @property
    def limit(self):
        """Number, ``np.nan`` or ``np.inf`` : Limit when `n` tends to infinity.

        Examples
        --------
            >>> Asymptotic(mu=-1, nu=0, xi=0).limit
            0
            >>> Asymptotic(mu=1, nu=0, xi=0).limit
            inf
            >>> Asymptotic(mu=-1, nu=np.nan, xi=np.nan).limit
            0
            >>> Asymptotic(mu=0, nu=-1, xi=np.nan).limit
            0

        ``np.nan`` means that the limit is unknown:

            >>> Asymptotic(mu=0, nu=0, xi=np.nan).limit
            nan
        """
        if isnan(self.mu):
            return np.nan
        elif self.mu > 0:
            return np.inf
        elif self.mu < 0:
            return 0
        else:
            # self.mu == 0
            if isnan(self.nu):
                return np.nan
            elif self.nu > 0:
                return np.inf
            elif self.nu < 0:
                return 0
            else:
                # self.nu == 0:
                if isnan(self.xi):
                    return np.nan
                else:
                    return exp(self.xi)

    def __mul__(self, other):
        """Multiplication of two asymptotic developments.

        Parameters
        ----------
        other: `Asymptotic` or Number.

        Returns
        -------
        product : Asymptotic
            The product of this asymptotic development and `other`.

        Examples
        --------
            >>> print(Asymptotic(mu=42, nu=51, xi=69) * Asymptotic(mu=1, nu=2, xi=3))
            exp(43 n + 53 log n + 72 + o(1))
            >>> print(Asymptotic(mu=42, nu=51, xi=69) * Asymptotic(mu=1, nu=np.nan, xi=np.nan))
            exp(43 n + ? log n + ? + o(1))
        """
        if not isinstance(other, Asymptotic):
            other = Asymptotic(0, 0, log(other))
        return Asymptotic(
            mu=0 if isclose(self.mu, -other.mu) else self.mu + other.mu,
            nu=0 if isclose(self.nu, -other.nu) else self.nu + other.nu,
            xi=0 if isclose(self.xi, -other.xi) else self.xi + other.xi
        )

    def __rmul__(self, other):
        """
        Examples
        --------
            >>> print(3 * Asymptotic(mu=42, nu=51, xi=69))
            exp(42 n + 51 log n + 70.0986 + o(1))
        """
        return self * other

    def __truediv__(self, other):
        """Division of two asymptotic developments.

        Parameters
        ----------
        other: `Asymptotic` or Number.

        Returns
        -------
        ratio : Asymptotic
            The division of this asymptotic development by `other`.

        Examples
        --------
            >>> print(Asymptotic(mu=42, nu=51, xi=69) / Asymptotic(mu=1, nu=2, xi=3))
            exp(41 n + 49 log n + 66 + o(1))
            >>> print(Asymptotic(mu=42, nu=51, xi=69) / Asymptotic(mu=1, nu=np.nan, xi=np.nan))
            exp(41 n + ? log n + ? + o(1))
            >>> print(1 / Asymptotic(mu=42, nu=51, xi=69))
            exp(- 42 n - 51 log n - 69 + o(1))
            >>> print(Asymptotic(mu=42, nu=51, xi=69) / 2)
            exp(42 n + 51 log n + 68.3069 + o(1))
        """
        if not isinstance(other, Asymptotic):
            other = Asymptotic(0, 0, log(other))
        return self * Asymptotic(- other.mu, - other.nu, - other.xi)

    def __rtruediv__(self, other):
        if not isinstance(other, Asymptotic):
            other = Asymptotic(0, 0, log(other))
        return other / self

    # noinspection PyTypeChecker
    def __add__(self, other):
        """Addition of two asymptotic developments.

        Parameters
        ----------
        other: `Asymptotic` or Number.

        Returns
        -------
        sum : Asymptotic
            The sum of this asymptotic development and `other`.

        Examples
        --------
            >>> print(Asymptotic(mu=42, nu=2, xi=69) + Asymptotic(mu=1, nu=51, xi=3))
            exp(42 n + 2 log n + 69 + o(1))
            >>> print(Asymptotic(mu=42, nu=2, xi=69) + Asymptotic(mu=42, nu=51, xi=3))
            exp(42 n + 51 log n + 3 + o(1))
            >>> print(Asymptotic(mu=42, nu=2, xi=4) + Asymptotic(mu=42, nu=2, xi=3))
            exp(42 n + 2 log n + 4.31326 + o(1))
            >>> print(Asymptotic(mu=42, nu=2, xi=4) + 1)
            exp(42 n + 2 log n + 4 + o(1))
            >>> print(1 + Asymptotic(mu=42, nu=2, xi=4))
            exp(42 n + 2 log n + 4 + o(1))
            >>> print(Asymptotic(mu=42, nu=2, xi=69) + Asymptotic(mu=41.99999999, nu=51, xi=3))
            exp(42 n + 51 log n + 3 + o(1))
        """
        if not isinstance(other, Asymptotic):
            other = Asymptotic(0, 0, log(other))
        if isnan(self.mu) or isnan(other.mu):
            return Asymptotic(np.nan, np.nan, np.nan)
        elif isclose(self.mu, other.mu):
            mu = max(self.mu, other.mu)
            if isnan(self.nu) or isnan(other.nu):
                return Asymptotic(mu, np.nan, np.nan)
            elif isclose(self.nu, other.nu):
                nu = max(self.nu, other.nu)
                xi = log(exp(self.xi) + exp(other.xi))
                return Asymptotic(mu, nu, xi)
            elif self.nu > other.nu:
                return Asymptotic(mu, self.nu, self.xi)
            else:
                return Asymptotic(mu, other.nu, other.xi)
        elif self.mu > other.mu:
            return Asymptotic(self.mu, self.nu, self.xi)
        else:
            return Asymptotic(other.mu, other.nu, other.xi)

    def __radd__(self, other):
        return self + other

    def isclose(self, other, *args, **kwargs):
        """Test near-equality.

        Parameters
        ----------
        other : Asymptotic
        *args
            Cf. :func:`math.isclose`.
        **kwargs
            Cf. :func:`math.isclose`.

        Returns
        -------
        isclose : bool
            True if this asymptotic development is approximately equal to `other`.

        Examples
        --------
            >>> Asymptotic(mu=1, nu=2, xi=3).isclose(
            ...     Asymptotic(mu=0.999999999999, nu=2.00000000001, xi=3))
            True
        """
        if not isinstance(other, Asymptotic):
            return False
        some_coefficients_are_nan = (isnan(self.mu) or isnan(self.nu) or isnan(self.xi)
                                     or isnan(other.mu) or isnan(other.nu) or isnan(other.xi))
        if some_coefficients_are_nan:
            raise ValueError('Can assert isclose only when all coefficients are known.')
        return (isclose(self.mu, other.mu, *args, **kwargs)
                and isclose(self.nu, other.nu, * args, ** kwargs)
                and isclose(self.xi, other.xi, *args, **kwargs))

    @classmethod
    def poisson_value(cls, tau, k):
        """Asymptotic development of ``P(X = k)``, where ``X ~ Poison(tau * n)``.

        Parameters
        ----------
        tau : Number
            The parameter of the Poisson distribution is ``tau * n``, where `n` tends to infinity.
        k : int
            The desired value in ``P(X = k)``.

        Returns
        -------
        Asymptotic
            The asymptotic development of ``P(X = k)``, where ``X ~ Poison(tau * n)``.

        Examples
        --------
            >>> print(Asymptotic.poisson_value(tau=0, k=0))
            exp(o(1))
            >>> print(Asymptotic.poisson_value(tau=0, k=1))
            exp(- inf)
            >>> print(Asymptotic.poisson_value(tau=1, k=1))
            exp(- n + log n + o(1))
            >>> from math import log
            >>> Asymptotic.poisson_value(tau=2, k=3).isclose(
            ...     Asymptotic(mu=-2, nu=3, xi=3 * log(2) - log(6)))
            True
        """
        if tau == 0:
            if k == 0:
                return cls(mu=0, nu=0, xi=0)
            else:
                return cls(mu=-np.inf, nu=-np.inf, xi=-np.inf)
        return cls(mu=- tau, nu=k, xi=k * log(tau) - log(factorial(k)))

    @classmethod
    def poisson_x1_eq_x2_plus_k(cls, tau_1, tau_2, k):
        """Asymptotic development of ``P(X_1 = X_2 + k)``, where ``X_i ~ Poison(tau_i * n)``.

        Parameters
        ----------
        tau_1,tau_2 : Number
            The parameter of the Poisson distribution of ``X_i`` is ``tau_i * n``, where `n` tends to infinity.
        k : int
            The desired value in ``P(X_1 = X_2 + k)``.

        Returns
        -------
        Asymptotic
            The asymptotic development of ``P(X_1 = X_2 + k)``, where ``X_i ~ Poison(tau_i * n)``.

        Examples
        --------
            >>> from math import log, pi, sqrt
            >>> print(Asymptotic.poisson_x1_eq_x2_plus_k(tau_1=0, tau_2=1, k=0))
            exp(- n + o(1))
            >>> print(Asymptotic.poisson_x1_eq_x2_plus_k(tau_1=0, tau_2=1, k=1))
            exp(- inf)
            >>> print(Asymptotic.poisson_x1_eq_x2_plus_k(tau_1=1, tau_2=0, k=1))
            exp(- n + log n + o(1))
            >>> Asymptotic.poisson_x1_eq_x2_plus_k(tau_1=2, tau_2=0, k=3).isclose(
            ...     Asymptotic(mu=-2, nu=3, xi=3 * log(2) - log(6)))
            True
            >>> Asymptotic.poisson_x1_eq_x2_plus_k(tau_1=1, tau_2=1, k=0).isclose(
            ...     Asymptotic(mu=0, nu=- 1 / 2, xi=- 1 / 2 * log(4 * pi)))
            True
            >>> Asymptotic.poisson_x1_eq_x2_plus_k(tau_1=1, tau_2=2, k=3).isclose(
            ...     Asymptotic(mu=- (1 - sqrt(2)) ** 2, nu=- 1 / 2, xi=- 1 / 2 * log(32 * pi * sqrt(2))))
            True
        """
        if tau_1 == 0:
            if k == 0:
                return cls(mu=- tau_2, nu=0, xi=0)
            else:
                return cls(mu=-np.inf, nu=-np.inf, xi=-np.inf)
        elif tau_2 == 0:
            return cls(mu=- tau_1, nu=k, xi=k * log(tau_1) - log(factorial(k)))
        elif tau_1 == tau_2:
            # Theoretically it falls into the general case. But in practice, the general case would give mu = -0.0,
            # which is not nice.
            return cls(
                mu=0,
                nu=- 1 / 2,
                xi=- 1 / 2 * log(4 * pi * tau_1)
            )
        else:
            return cls(
                mu=- (sqrt(tau_1) - sqrt(tau_2)) ** 2,
                nu=- 1 / 2,
                xi=- 1 / 2 * log(4 * pi * sqrt(tau_1 * tau_2) * tau_2**k / tau_1**k)
            )

    @classmethod
    def poisson_eq(cls, tau_1, tau_2):
        """Asymptotic development of ``P(X_1 = X_2)``, where ``X_i ~ Poison(tau_i * n)``.

        Parameters
        ----------
        tau_1,tau_2 : Number
            The parameter of the Poisson distribution of ``X_i`` is ``tau_i * n``, where `n` tends to infinity.

        Returns
        -------
        Asymptotic
            The asymptotic development of ``P(X_1 = X_2)``, where ``X_i ~ Poison(tau_i * n)``.

        Examples
        --------
            >>> from math import pi, log
            >>> print(Asymptotic.poisson_eq(tau_1=0, tau_2=0))
            exp(o(1))
            >>> print(Asymptotic.poisson_eq(tau_1=1/10, tau_2=0))
            exp(- 0.1 n + o(1))
            >>> Asymptotic.poisson_eq(tau_1=1/10, tau_2=9/10).isclose(
            ...     Asymptotic(mu=-.4, nu=-.5, xi=-.5 * log(1.2 * pi)))
            True
        """
        return cls.poisson_x1_eq_x2_plus_k(tau_1, tau_2, 0)

    @classmethod
    def poisson_one_more(cls, tau_1, tau_2):
        """Asymptotic development of ``P(X_1 = X_2 + 1)``, where ``X_i ~ Poison(tau_i * n)``.

        Parameters
        ----------
        tau_1,tau_2 : Number
            The parameter of the Poisson distribution of ``X_i`` is ``tau_i * n``, where `n` tends to infinity.

        Returns
        -------
        Asymptotic
            The asymptotic development of ``P(X_1 = X_2 + 1)``, where ``X_i ~ Poison(tau_i * n)``.

        Examples
        --------
            >>> from math import pi, log
            >>> print(Asymptotic.poisson_one_more(tau_1=0, tau_2=1/10))
            exp(- inf)
            >>> print(Asymptotic.poisson_one_more(tau_1=1/10, tau_2=0))
            exp(- 0.1 n + log n - 2.30259 + o(1))

        """
        return cls.poisson_x1_eq_x2_plus_k(tau_1, tau_2, 1)

    @classmethod
    def poisson_x1_ge_x2_plus_k(cls, tau_1, tau_2, k):
        """Asymptotic development of ``P(X_1 >= X_2 + k)``, where ``X_i ~ Poison(tau_i * n)``.

        Parameters
        ----------
        tau_1,tau_2 : Number
            The parameter of the Poisson distribution of ``X_i`` is ``tau_i * n``, where `n` tends to infinity.
        k : int
            The desired value in ``P(X_1 >= X_2 + k)``.

        Returns
        -------
        Asymptotic
            The asymptotic development of ``P(X_1 >= X_2 + k)``, where ``X_i ~ Poison(tau_i * n)``.

        Examples
        --------
            >>> from math import log
            >>> print(Asymptotic.poisson_x1_ge_x2_plus_k(tau_1=0, tau_2=1, k=0))
            exp(- n + o(1))
            >>> print(Asymptotic.poisson_x1_ge_x2_plus_k(tau_1=0, tau_2=1, k=1))
            exp(- inf)
            >>> print(Asymptotic.poisson_x1_ge_x2_plus_k(tau_1=1, tau_2=0, k=0))
            exp(o(1))
            >>> Asymptotic.poisson_x1_ge_x2_plus_k(tau_1=1, tau_2=1, k=0).isclose(
            ...     Asymptotic(mu=0, nu=0, xi=- log(2)))
            True
            >>> Asymptotic.poisson_x1_ge_x2_plus_k(tau_1=1, tau_2=2, k=3).isclose(Asymptotic(
            ...     mu=- (1 - sqrt(2)) ** 2, nu=- 1 / 2,
            ...     xi=(- 1 / 2 * log(32 * pi * sqrt(2)) - log(1 - sqrt(1 / 2)))))
            True
        """
        if tau_1 == 0:
            if k == 0:
                return cls(mu=- tau_2, nu=0, xi=0)
            else:
                return cls(mu=-np.inf, nu=-np.inf, xi=-np.inf)
        elif tau_1 > tau_2:
            # Probability 1 => the log tends to 0.
            return cls(mu=0, nu=0, xi=0)
        elif tau_1 == tau_2:
            # Probability 1/2 => the log tends to - log(2).
            return cls(mu=0, nu=0, xi=- log(2))
        else:
            # Use the offset theorem with event X_1 = X_2, then infinite sum.
            return cls(
                mu=- (sqrt(tau_1) - sqrt(tau_2)) ** 2,
                nu=- 1 / 2,
                xi=(- 1 / 2 * log(4 * pi * sqrt(tau_1 * tau_2) * tau_2**k / tau_1**k)
                    - log(1 - sqrt(tau_1 / tau_2)))
            )

    @classmethod
    def poisson_ge(cls, tau_1, tau_2):
        """Asymptotic development of ``P(X_1 >= X_2)``, where ``X_i ~ Poison(tau_i * n)``.

        Parameters
        ----------
        tau_1,tau_2 : Number
            The parameter of the Poisson distribution of ``X_i`` is ``tau_i * n``, where `n` tends to infinity.

        Returns
        -------
        Asymptotic
            The asymptotic development of ``P(X_1 >= X_2)``, where ``X_i ~ Poison(tau_i * n)``.

        Examples
        --------
            >>> from math import log
            >>> print(Asymptotic.poisson_ge(tau_1=0, tau_2=0))
            exp(o(1))
            >>> print(Asymptotic.poisson_ge(tau_1=0, tau_2=1/10))
            exp(- 0.1 n + o(1))
            >>> print(Asymptotic.poisson_ge(tau_1=1/10, tau_2=0))
            exp(o(1))
            >>> Asymptotic.poisson_ge(tau_1=9/10, tau_2=1/10).isclose(
            ...     Asymptotic(mu=0, nu=0, xi=0))
            True
            >>> Asymptotic.poisson_ge(tau_1=3/10, tau_2=3/10).isclose(
            ...     Asymptotic(mu=0, nu=0, xi=-log(2)))
            True
            >>> asymptotic = Asymptotic.poisson_ge(tau_1=1/10, tau_2=9/10)
            >>> asymptotic_eq = Asymptotic.poisson_eq(tau_1=1/10, tau_2=9/10)
            >>> asymptotic.isclose(
            ...     Asymptotic(mu=asymptotic_eq.mu, nu=asymptotic_eq.nu, xi=asymptotic_eq.xi + log(3/2)))
            True
        """
        return cls.poisson_x1_ge_x2_plus_k(tau_1, tau_2, 0)

    @classmethod
    def poisson_gt(cls, tau_1, tau_2):
        """Asymptotic development of ``P(X_1 > X_2)``, where ``X_i ~ Poison(tau_i * n)``.

        Parameters
        ----------
        tau_1,tau_2 : Number
            The parameter of the Poisson distribution of ``X_i`` is ``tau_i * n``, where `n` tends to infinity.

        Returns
        -------
        Asymptotic
            The asymptotic development of ``P(X_1 > X_2)``, where ``X_i ~ Poison(tau_i * n)``.

        Examples
        --------
            >>> from math import log
            >>> print(Asymptotic.poisson_gt(tau_1=0, tau_2=0))
            exp(- inf)
            >>> print(Asymptotic.poisson_gt(tau_1=0, tau_2=1/10))
            exp(- inf)
            >>> print(Asymptotic.poisson_gt(tau_1=1/10, tau_2=0))
            exp(o(1))
            >>> Asymptotic.poisson_gt(tau_1=9/10, tau_2=1/10).isclose(
            ...     Asymptotic(mu=0, nu=0, xi=0))
            True
            >>> Asymptotic.poisson_gt(tau_1=3/10, tau_2=3/10).isclose(
            ...     Asymptotic(mu=0, nu=0, xi=-log(2)))
            True
            >>> asymptotic = Asymptotic.poisson_gt(tau_1=1/10, tau_2=9/10)
            >>> asymptotic_eq = Asymptotic.poisson_eq(tau_1=1/10, tau_2=9/10)
            >>> asymptotic.isclose(
            ...     Asymptotic(mu=asymptotic_eq.mu, nu=asymptotic_eq.nu, xi=asymptotic_eq.xi - log(2)))
            True
        """
        return cls.poisson_x1_ge_x2_plus_k(tau_1, tau_2, 1)

    @classmethod
    def poisson_gt_one_more(cls, tau_1, tau_2):
        """Asymptotic development of ``P(X_1 > X_2 + 1)``, where ``X_i ~ Poison(tau_i * n)``.

        Parameters
        ----------
        tau_1,tau_2 : Number
            The parameter of the Poisson distribution of ``X_i`` is ``tau_i * n``, where `n` tends to infinity.

        Returns
        -------
        Asymptotic
            The asymptotic development of ``P(X_1 > X_2 + 1)``, where ``X_i ~ Poison(tau_i * n)``.

        Examples
        --------
            >>> from math import log
            >>> print(Asymptotic.poisson_gt_one_more(tau_1=0, tau_2=0))
            exp(- inf)
            >>> print(Asymptotic.poisson_gt_one_more(tau_1=0, tau_2=1/10))
            exp(- inf)
            >>> print(Asymptotic.poisson_gt_one_more(tau_1=1/10, tau_2=0))
            exp(o(1))
            >>> Asymptotic.poisson_gt_one_more(tau_1=9/10, tau_2=1/10).isclose(
            ...     Asymptotic(mu=0, nu=0, xi=0))
            True
            >>> Asymptotic.poisson_gt_one_more(tau_1=3/10, tau_2=3/10).isclose(
            ...     Asymptotic(mu=0, nu=0, xi=-log(2)))
            True
            >>> asymptotic = Asymptotic.poisson_gt_one_more(tau_1=1/10, tau_2=9/10)
            >>> asymptotic_eq = Asymptotic.poisson_eq(tau_1=1/10, tau_2=9/10)
            >>> asymptotic.isclose(
            ...     Asymptotic(mu=asymptotic_eq.mu, nu=asymptotic_eq.nu, xi=asymptotic_eq.xi - log(6)))
            True
        """
        return cls.poisson_x1_ge_x2_plus_k(tau_1, tau_2, 2)



import numpy as np
from poisson_approval.utils.Util import isneginf
from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event
from poisson_approval.events.EventTrio import EventTrio


class EventTrio2t(Event):
    """A 3-candidate almost-tie of the type (X - 1, X - 1, X).

    Notes
    -----
    We consider the trios of type 1, i.e. situations where ``S_x + 1 = S_y + 1 = S_z``.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> EventTrio2t(candidate_x='a', candidate_y='b', candidate_z='c',
        ...             tau_a=Fraction(1, 10), tau_ab=Fraction(6, 10), tau_c=Fraction(3, 10))
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 1.18339 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        event_trio = EventTrio(candidate_x='x', candidate_y='y', candidate_z='z',
                               tau_x=tau_x, tau_y=tau_y, tau_z=tau_z, tau_xy=tau_xy, tau_xz=tau_xz, tau_yz=tau_yz)
        self._phi_x = event_trio.phi['x']
        self._phi_y = event_trio.phi['y']
        self._phi_z = event_trio.phi['z']
        self._phi_xy = event_trio.phi['xy']
        self._phi_xz = event_trio.phi['xz']
        self._phi_yz = event_trio.phi['yz']
        if tau_z == 0 and tau_xy == 0:
            # Cross diagram 1
            self.asymptotic = Asymptotic.poisson_one_more(tau_yz, tau_x) * Asymptotic.poisson_one_more(tau_xz, tau_y)
        elif (tau_x == 0 and tau_yz == 0) or (tau_y == 0 and tau_xz == 0):
            # Cross diagram 2
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, tau_yz) * Asymptotic.poisson_eq(tau_y, tau_xz)
                               * Asymptotic.poisson_one_more(tau_z, tau_xy))
        elif (tau_z == 0 and tau_xz == 0) or (tau_z == 0 and tau_yz == 0):
            # Flower diagram 1
            self.asymptotic = Asymptotic(mu=-np.inf, nu=-np.inf, xi=-np.inf)
        elif (tau_x == 0 and tau_xz == 0) or (tau_y == 0 and tau_yz == 0):
            # Flower diagram 2
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, tau_yz) * Asymptotic.poisson_eq(tau_y, tau_xz)
                               * Asymptotic.poisson_one_more(tau_z, tau_xy))
        elif (tau_x == 0 and tau_xy == 0) or (tau_y == 0 and tau_xy == 0):
            # Flower diagram 3
            self.asymptotic = (
                (Asymptotic.poisson_value(tau_z, 0)
                 * Asymptotic.poisson_one_more(tau_yz, tau_x) * Asymptotic.poisson_one_more(tau_xz, tau_y))
                + (Asymptotic.poisson_one_more(tau_z, 1)
                   * Asymptotic.poisson_eq(tau_x, tau_yz) * Asymptotic.poisson_eq(tau_y, tau_xz))
            )
        else:
            _phi_xy_tilde = self._phi_xy if tau_xy != 0 else self._phi_x * self._phi_y
            self.asymptotic = event_trio.asymptotic * _phi_xy_tilde
        if isneginf(self.asymptotic.mu):
            self._phi_x = np.nan
            self._phi_y = np.nan
            self._phi_z = np.nan
            self._phi_xy = np.nan
            self._phi_xz = np.nan
            self._phi_yz = np.nan



import math
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal
from poisson_approval.utils.Util import give_figure
from fractions import Fraction


class ExploreGridProfilesOrdinal:
    """Explore a grid of ordinal profiles and analyze each profile.

    Parameters
    ----------
    denominator : int or list of int
        The grain of the grid. For example, if ``denominator=9``, we examine profiles such as
        (1/9, 2/9, 2/9, 1/9, 1/3, 0) or (1/9, 5/9, 0, 3/9, 0, 0), etc. If a list of integers is given, then all the
        corresponding denominators are examined.
    test : callable
        A function ``ProfileOrdinal -> bool``. Only the profiles satisfying this test will be examined.
        Default: always True.
    standardized : bool
        If True, then we examine only the `standardized` profiles (i.e. up to relabelling the candidates).
        Cf. :attr:`Profile.is_standardized`.
    well_informed_voters : bool
        Cf. the corresponding parameter in :class:`Profile`.

    Examples
    --------
        >>> def test(profile):
        ...     return(
        ...         profile.is_profile_condorcet
        ...         and len(profile.analyzed_strategies.utility_dependent) > 0
        ...     )
        >>> exploration = ExploreGridProfilesOrdinal(denominator=3, test=test)
        >>> exploration
        0 equilibrium, 1 utility-dependent equilibrium, 3 non-equilibria
        <abc: 2/3, cab: 1/3> (Condorcet winner: a)
        <BLANKLINE>
        0 equilibrium, 1 utility-dependent equilibrium, 7 non-equilibria
        <abc: 1/3, acb: 1/3, bac: 1/3> (Condorcet winner: a)
        <BLANKLINE>
        0 equilibrium, 2 utility-dependent equilibria, 2 non-equilibria
        <abc: 2/3, cba: 1/3> (Condorcet winner: a)
        <abc: 2/3, bca: 1/3> (Condorcet winner: a)
        <BLANKLINE>
        0 equilibrium, 3 utility-dependent equilibria, 5 non-equilibria
        <abc: 1/3, acb: 1/3, bca: 1/3> (Condorcet winner: a)
        <BLANKLINE>
        2 equilibria, 1 utility-dependent equilibrium, 5 non-equilibria
        <abc: 1/3, bac: 1/3, cab: 1/3> (Condorcet winner: a)
        <BLANKLINE>
        >>> profile = exploration[(0, 1, 3)][0]
        >>> print(profile)
        <abc: 2/3, cab: 1/3> (Condorcet winner: a)

    If the test breaks the symmetry between candidates, you should use the option ``𝚜𝚝𝚊𝚗𝚍𝚊𝚛𝚍𝚒𝚣𝚎𝚍=𝙵𝚊𝚕𝚜𝚎``:

        >>> def test(profile):
        ...     return (
        ...         'a' in profile.condorcet_winners
        ...     )
        >>> exploration = ExploreGridProfilesOrdinal(test=test, denominator=2, standardized=False)
        >>> exploration
        0 equilibrium, 1 utility-dependent equilibrium, 3 non-equilibria
        <bac: 1/2, cab: 1/2> (Condorcet winner: a, b, c)
        <acb: 1/2, cba: 1/2> (Condorcet winner: a, c)
        <acb: 1/2, bca: 1/2> (Condorcet winner: a, b, c)
        <acb: 1/2, bac: 1/2> (Condorcet winner: a, b)
        <abc: 1/2, cba: 1/2> (Condorcet winner: a, b, c)
        <abc: 1/2, cab: 1/2> (Condorcet winner: a, c)
        <abc: 1/2, bca: 1/2> (Condorcet winner: a, b)
        <BLANKLINE>
        1 equilibrium, 0 utility-dependent equilibrium, 1 non-equilibrium
        <acb: 1> (Condorcet winner: a)
        <abc: 1> (Condorcet winner: a)
        <BLANKLINE>
        1 equilibrium, 0 utility-dependent equilibrium, 3 non-equilibria
        <acb: 1/2, cab: 1/2> (Condorcet winner: a, c)
        <abc: 1/2, bac: 1/2> (Condorcet winner: a, b)
        <abc: 1/2, acb: 1/2> (Condorcet winner: a)
        <BLANKLINE>
    """

    def __init__(self, denominator, test=None, standardized=True, well_informed_voters=True):
        if type(denominator) == int:
            denominators = [denominator]
        else:
            denominators = denominator
        if test is None:
            def test(_):
                return True
        self.d_stats_profiles = dict()
        for d in denominators:
            r_abc_min = int(math.ceil(d / 6)) if standardized else 0
            for r_abc in range(r_abc_min, d + 1):
                r_acb_max = d - r_abc
                if standardized:
                    r_acb_max = min(r_abc, r_acb_max)
                for r_acb in range(0, r_acb_max + 1):
                    r_bac_max = d - r_abc - r_acb
                    if standardized:
                        r_bac_max = min(r_abc, r_bac_max)
                    for r_bac in range(0, r_bac_max + 1):
                        r_bca_max = d - r_abc - r_acb - r_bac
                        if standardized:
                            r_bca_max = min(r_abc, r_bca_max)
                        for r_bca in range(0, r_bca_max + 1):
                            r_cab_max = d - r_abc - r_acb - r_bac - r_bca
                            if standardized:
                                r_cab_max = min(r_abc, r_cab_max)
                            for r_cab in range(0, r_cab_max + 1):
                                r_cba = (
                                    d - r_abc - r_acb - r_bac - r_bca - r_cab)
                                profile = ProfileOrdinal({
                                    'abc': Fraction(r_abc, d),
                                    'acb': Fraction(r_acb, d),
                                    'bac': Fraction(r_bac, d),
                                    'bca': Fraction(r_bca, d),
                                    'cab': Fraction(r_cab, d),
                                    'cba': Fraction(r_cba, d)
                                }, well_informed_voters=well_informed_voters)
                                if standardized and not profile.is_standardized:
                                    continue
                                if not test(profile):
                                    continue
                                analyzed_strat = profile.analyzed_strategies
                                eq, ud, non_eq = (
                                    analyzed_strat.equilibria,
                                    analyzed_strat.utility_dependent,
                                    analyzed_strat.non_equilibria
                                )
                                stats = (len(eq), len(ud), len(non_eq))
                                if stats not in self.d_stats_profiles:
                                    self.d_stats_profiles[stats] = []
                                if profile not in self.d_stats_profiles[stats]:
                                    self.d_stats_profiles[stats].append(profile)

    def __getitem__(self, item):
        return self.d_stats_profiles[item]

    def items(self):
        return self.d_stats_profiles.items()

    def keys(self):
        return self.d_stats_profiles.keys()

    def values(self):
        return self.d_stats_profiles.values()

    def __repr__(self):
        s = ''
        first = True
        for key in sorted(self.d_stats_profiles):
            list_profiles = self.d_stats_profiles[key]
            if first:
                first = False
            else:
                s += '\n'
            s += give_figure(key[0], 'equilibrium', 'equilibria') + ', '
            s += give_figure(key[1], 'utility-dependent equilibrium', 'utility-dependent equilibria') + ', '
            s += give_figure(key[2], 'non-equilibrium', 'non-equilibria') + '\n'
            for profile in list_profiles:
                s += str(profile) + '\n'
        return s



from poisson_approval.utils.Util import initialize_random_seeds


class GeneratorExample:
    """A generator of `something` that meets a given test.

    Parameters
    ----------
    generator : callable or tuple
        This can be:

        * Either a callable that takes no input and that outputs a (random) `something`,
        * Or a tuple of such generators (cf. examples below).
    test : callable
        A function that take as input(s) the output(s) of the generator(s) and that returns a Boolean.
    n_trials_max : int or None
        The maximum number of trials. If None, then attempts will be made until finding an example. The option None
        should be used with extreme care since it may lead to an infinite loop.

    Returns
    -------
    object or tuple or None
        If there is one generator, then return a `something` that meets the given `test`. If there is a tuple of
        generators, then return a tuple of generated objects that meets the given `test`. In both cases, if no example
        is found within `n_trials_max` attempts, then return None.

    Examples
    --------
    In this basic example with one generator, we generate a random integer between 0 included and 100 excluded that
    is divisible by 7:

        >>> import numpy as np
        >>> initialize_random_seeds()
        >>> def generator_integer():
        ...     return np.random.randint(0, 100)
        >>> def test_divisible_7(n):
        ...     return n % 7 == 0
        >>> generator_example = GeneratorExample(generator=generator_integer,
        ...                                      test=test_divisible_7, n_trials_max=None)
        >>> generator_example()
        21

    In this example with a tuple of generators, we generate a random 2*2 matrix and a random vector of size 2, both
    with integer coefficients between -10 included and 11 excluded, such that their dot product is null, but without
    being null themselves:

        >>> import numpy as np
        >>> initialize_random_seeds()
        >>> def generator_matrix():
        ...     return np.random.randint(-10, 11, (2, 2))
        >>> def generator_vector():
        ...     return np.random.randint(-10, 11, 2)
        >>> def test_non_trivial_dot_zero(matrix, vector):
        ...     return (np.all(np.dot(matrix, vector) == 0)
        ...             and not np.all(matrix == 0) and not np.all(vector == 0))
        >>> generator_example = GeneratorExample(generator=(generator_matrix, generator_vector),
        ...                                      test=test_non_trivial_dot_zero, n_trials_max=None)
        >>> matrix, vector = generator_example()
        >>> matrix
        array([[ 2, -8],
               [-1,  4]])
        >>> vector
        array([-4, -1])

    When no example is found, the generator returns None:

        >>> import numpy as np
        >>> def generator_integer():
        ...     return np.random.randint(0, 100)
        >>> def is_negative(n):
        ...     return n < 0
        >>> generator_example = GeneratorExample(generator=generator_integer,
        ...                                      test=is_negative, n_trials_max=100)
        >>> n = generator_example()
        >>> print(n)
        None
    """

    def __init__(self, generator, test, n_trials_max):
        self.generator = generator
        self.test = test
        self.n_trials_max = n_trials_max

    def __call__(self):
        i = 0
        while self.n_trials_max is None or i < self.n_trials_max:
            i += 1
            if isinstance(self.generator, tuple):
                somethings = [g() for g in self.generator]
                if self.test(*somethings):
                    return tuple(somethings)
            else:
                something = self.generator()
                if self.test(something):
                    return something
        return None



from poisson_approval import StrategyOrdinal, PLURALITY, ANTI_PLURALITY


def test_strategy_ordinal_plurality():
    """
        >>> strategy = StrategyOrdinal({'abc': 'a', 'cba': 'b'}, voting_rule=PLURALITY)
        >>> strategy
        StrategyOrdinal({'abc': 'a', 'cba': 'b'}, voting_rule='Plurality')
        >>> print(strategy)
        <abc: a, cba: b> (Plurality)
    """
    pass


def test_strategy_ordinal_anti_plurality():
    """
        >>> strategy = StrategyOrdinal({'abc': 'ab', 'cba': 'ac'}, voting_rule=ANTI_PLURALITY)
        >>> strategy
        StrategyOrdinal({'abc': 'ab', 'cba': 'ac'}, voting_rule='Anti-plurality')
        >>> print(strategy)
        <abc: ab, cba: ac> (Anti-plurality)
    """
    pass



from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex
from poisson_approval.tau_vector.TauVector import TauVector


class GeneratorTauVectorUniform:
    """A generator of tau-vectors (:class:`TauVector`) following the uniform distribution.

    Parameters
    ----------
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`TauVector`.

    Notes
    -----
    The tau-vector is drawn uniformly on the simplex.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> generator = GeneratorTauVectorUniform()
        >>> tau = generator()
        >>> print(tau)
        <a: 0.4236547993389047, ab: 0.05394987214431912, ac: 0.1124259903007756, b: 0.12122838365799216, \
bc: 0.2848106336275805, c: 0.0039303209304278885> ==> a
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self):
        """
        Returns
        -------
        TauVector
            A tau-vector.
        """
        x = rand_simplex(d=6)
        return TauVector({ballot: x[i] for i, ballot in enumerate(BALLOTS_WITHOUT_INVERSIONS)}, **self.kwargs)



import random
import numpy as np
from math import isclose
from itertools import product


def masks_area_naive(inf, sup, masks):
    """Area of some masks. Naive implementation (used as a sanity test for the recursive implementation)

    Notes
    -----
    Same specifications as :meth:`masks_area`.

    Examples
    --------
        >>> area = masks_area_naive(inf=[0, 0], sup=[1, 2],
        ...                         masks=[[(.2, True), (.6, False)], [(.3, False), (.4, True)]])
        >>> isclose(area, 0.8 * 0.6 + 0.3 * 1.6 - 0.1 * 0.2)
        True
    """
    dim = len(inf)
    limits = [sorted({mask[d][0] for mask in masks}.union({inf[d], sup[d]})) for d in range(dim)]
    medians_and_lengths = [
        [((limits[d][i + 1] + limits[d][i]) / 2, limits[d][i + 1] - limits[d][i]) for i in range(len(limits[d]) - 1)]
        for d in range(dim)]
    area = 0
    for point in product(*medians_and_lengths):
        if any([all([(point[d][0] > mask[d][0]) == mask[d][1] for d in range(dim)]) for mask in masks]):
            area += np.prod([point[d][1] for d in range(dim)])
    return area


def masks_area(inf, sup, masks):
    """Area of some masks (recursive `divide and conquer` implementation).

    We denote by `d` the dimension of the Euclidean space under study.

    Parameters
    ----------
    inf : list of Number
        A list of `d` numbers. The inf limit of the bounding rectangle in each dimension.
    sup : list of Number
        A list of `d` numbers. The sup limit of the bounding rectangle in each dimension.
    masks : list of list of tuple
        A list of masks. Each mask is a list of `d` pairs ``(lim, direction)``, where ``lim`` is the limit
        of the mask, and ``direction`` is a Boolean: ``True`` (resp ``False``) means that the points of the mask
        meet the condition ``x_d >= lim`` (resp. ``x_d <= lim``).

    Returns
    -------
    float
        The area of ``Intersection(bounding rectangle, Union(masks))``.

    Examples
    --------
    In the following example, the bounding rectangle is the set of points where ``0 <= x_1 <= 1 and 0 <= x_2 <= 2``.
    The first mask is the set of points where ``x_1 >= 0.2 and x_2 <= 0.6``. The second mask is the set of points
    where ``x_1 <= 0.3 and x_2 >= 0.4``.

        >>> area = masks_area(inf=[0, 0], sup=[1, 2], masks=[[(.2, True), (.6, False)], [(.3, False), (.4, True)]])
        >>> isclose(area, 0.8 * 0.6 + 0.3 * 1.6 - 0.1 * 0.2)
        True
    """
    dim = len(inf)
    # Is there a mask covering everything?
    if any([all([(mask[d][0] <= inf[d]) if mask[d][1] else (mask[d][0] >= sup[d]) for d in range(dim)])
            for mask in masks]):
        return np.prod([high - low for (low, high) in zip(inf, sup)])
    # Eliminate empty masks
    new_masks = [mask for mask in masks
                 if not any([(mask[d][0] >= sup[d]) if mask[d][1] else (mask[d][0] <= inf[d]) for d in range(dim)])]
    if not new_masks:
        return 0
    mask = random.choice(new_masks)
    d_lim = random.choice([d for d in range(dim) if inf[d] < mask[d][0] < sup[d]])
    lim = mask[d_lim][0]
    return (masks_area(inf, [lim if d == d_lim else sup[d] for d in range(dim)], new_masks)
            + masks_area([lim if d == d_lim else inf[d] for d in range(dim)], sup, new_masks))


def masks_distribution_naive(inf, sup, masks):
    """Distribution of the number of masks. Naive implementation (used as a sanity test for the recursive
    implementation)

    Notes
    -----
    Same specifications as :meth:`mask_distribution`.

    Examples
    --------
        >>> histogram = masks_distribution_naive(inf=[0, 0], sup=[1, 2],
        ...                                      masks=[[(.2, True), (.6, False)], [(.3, False), (.4, True)]])
        >>> histogram
        array([1.06, 0.92, 0.02])
    """
    dim = len(inf)
    limits = [sorted({mask[d][0] for mask in masks}.union({inf[d], sup[d]})) for d in range(dim)]
    medians_and_lengths = [
        [((limits[d][i + 1] + limits[d][i]) / 2, limits[d][i + 1] - limits[d][i]) for i in range(len(limits[d]) - 1)]
        for d in range(dim)]
    histogram = np.zeros(dim + 1)
    for point in product(*medians_and_lengths):
        n_masks = np.sum([all([(point[d][0] > mask[d][0]) == mask[d][1] for d in range(dim)]) for mask in masks])
        area = np.prod([point[d][1] for d in range(dim)])
        histogram[n_masks] += area
    return histogram


def masks_distribution(inf, sup, masks, cover_alls=0):
    """Distribution of the number of masks (recursive `divide and conquer` implementation).

    We denote by `d` the dimension of the Euclidean space under study.

    Parameters
    ----------
    inf : list of Number
        A list of `d` numbers. The inf limit of the bounding rectangle in each dimension.
    sup : list of Number
        A list of `d` numbers. The sup limit of the bounding rectangle in each dimension.
    masks : list of list of tuple
        A list of masks. Each mask is a list of `d` pairs ``(lim, direction)``, where ``lim`` is the limit
        of the mask, and ``direction`` is a Boolean: ``True`` (resp ``False``) means that the points of the mask
        meet the condition ``x_d >= lim`` (resp. ``x_d <= lim``).
    cover_alls : int, optional
        If specified, then we consider that we have this number of implicit masks (i.e. not given in the argument
        `masks`) that cover the whole area.

    Returns
    -------
    list
        A list. The `i`-th coefficient is the area covered by `i` masks exactly (and in the bounding rectangle).

    Examples
    --------
    In the following example, the bounding rectangle is the set of points where ``0 <= x_1 <= 1 and 0 <= x_2 <= 2``.
    The first mask is the set of points where ``x_1 >= 0.2 and x_2 <= 0.6``. The second mask is the set of points
    where ``x_1 <= 0.3 and x_2 >= 0.4``.

        >>> histogram = masks_distribution(inf=[0, 0], sup=[1, 2],
        ...                                masks=[[(.2, True), (.6, False)], [(.3, False), (.4, True)]])
        >>> histogram
        array([1.06, 0.92, 0.02])
    """
    result = masks_distribution_aux(inf, sup, masks, histogram=None, cover_alls=cover_alls)
    last_non_zero = result.size - 1
    while result[last_non_zero] == 0:
        last_non_zero -= 1
    return result[:last_non_zero + 1]


def masks_distribution_aux(inf, sup, masks, histogram=None, cover_alls=0):
    """Distribution of the number of masks (recursive `divide and conquer` implementation).

    We denote by `d` the dimension of the Euclidean space under study.

    Parameters
    ----------
    inf : list of Number
        A list of `d` numbers. The inf limit of the bounding rectangle in each dimension.
    sup : list of Number
        A list of `d` numbers. The sup limit of the bounding rectangle in each dimension.
    masks : list of list of tuple
        A list of masks. Each mask is a list of `d` pairs ``(lim, direction)``, where ``lim`` is the limit
        of the mask, and ``direction`` is a Boolean: ``True`` (resp ``False``) means that the points of the mask
        meet the condition ``x_d >= lim`` (resp. ``x_d <= lim``).
    histogram : list, optional
        This parameter should only be used for recursive calls. If specified, then instead of creating a new list for
        the output, it is added to the given list `histogram`.
    cover_alls : int, optional
        This parameter should only be used for recursive calls. If specified, then we consider that we have this number
        of implicit masks (i.e. not given in the argument `masks`) that cover the whole area.

    Returns
    -------
    list
        A list of length `d + 1`. The `i`-th coefficient is the area covered by `i` masks exactly (and in the bounding
        rectangle).
    """
    dim = len(inf)
    if histogram is None:
        histogram = np.zeros(dim + 1)
    # Eliminate empty masks
    new_masks = [mask for mask in masks
                 if not any([(mask[d][0] >= sup[d]) if mask[d][1] else (mask[d][0] <= inf[d]) for d in range(dim)])]
    # Detect the new cover-alls (masks covering everything)
    cover_alls += np.sum([all([(mask[d][0] <= inf[d]) if mask[d][1] else (mask[d][0] >= sup[d]) for d in range(dim)])
                          for mask in new_masks], dtype=int)
    new_masks = [mask for mask in new_masks
                 if not all([(mask[d][0] <= inf[d]) if mask[d][1] else (mask[d][0] >= sup[d]) for d in range(dim)])]
    if not new_masks:
        area = np.prod([high - low for (low, high) in zip(inf, sup)])
        histogram[cover_alls] += area
    else:
        mask = random.choice(new_masks)
        d_lim = random.choice([d for d in range(dim) if inf[d] < mask[d][0] < sup[d]])
        lim = mask[d_lim][0]
        masks_distribution_aux(inf, [lim if d == d_lim else sup[d] for d in range(dim)], new_masks,
                               histogram, cover_alls)
        masks_distribution_aux([lim if d == d_lim else inf[d] for d in range(dim)], sup, new_masks,
                               histogram, cover_alls)
    return histogram


def winners_distribution(inf, sup, masks_winners, histogram=None, cover_alls=None):
    """Distribution of the number of winners (recursive `divide and conquer` implementation).

    We denote by `d` the dimension of the Euclidean space under study.

    Parameters
    ----------
    inf : list of Number
        A list of `d` numbers. The inf limit of the bounding rectangle in each dimension.
    sup : list of Number
        A list of `d` numbers. The sup limit of the bounding rectangle in each dimension.
    masks_winners : list of tuple
        A list of pairs ``(mask, winners)``. A mask is defined as usual (cf. :meth:`masks_area` for instance).
        A winner is a set of winning candidates in this mask, e.g. ``{'a', 'b'}``.
    histogram : list
        This parameter should only be used for recursive calls. If specified, then instead of creating a new list for
        the output, it is added to the given list `histogram`.
    cover_alls : set
        E.g. {'a', 'b'}. This parameter should only be used for recursive calls. If specified, then we consider that
        we have all these candidates winning in the whole area.

    Returns
    -------
    list
        A list of length 4. The `i`-th coefficient is the area where `i` candidates may win.

    Examples
    --------
    In the following example, the bounding rectangle is the set of points where ``0 <= x_1 <= 1 and 0 <= x_2 <= 2``.
    The first mask is the set of points where ``x_1 >= 0.2 and x_2 <= 0.6``, where the winner can be ``{'a'}``.
    The second mask is the set of points where ``x_1 <= 0.3 and x_2 >= 0.4``, where the winner can be ``{'b'}``.

        >>> histogram = winners_distribution(
        ...     inf=[0, 0], sup=[1, 2],
        ...     masks_winners=[([(.2, True), (.6, False)], {'a'}), ([(.3, False), (.4, True)], {'a', 'b'})])
        >>> histogram
        array([1.06, 0.46, 0.48, 0.  ])
    """
    dim = len(inf)
    if histogram is None:
        histogram = np.zeros(4)
    if cover_alls is None:
        cover_alls = set()
    # Eliminate empty masks
    new_mw = [(mask, winners) for mask, winners in masks_winners
              if not any([(mask[d][0] >= sup[d]) if mask[d][1] else (mask[d][0] <= inf[d]) for d in range(dim)])
              and not winners.issubset(cover_alls)]
    # Detect the new cover-alls (masks covering everything)
    toto = set.union(*([set()] + [
        winners
        for mask, winners in new_mw
        if all([(mask[d][0] <= inf[d]) if mask[d][1] else (mask[d][0] >= sup[d]) for d in range(dim)])
    ]))
    cover_alls = cover_alls.union(toto)
    new_mw = [(mask, winners) for mask, winners in new_mw
              if not all([(mask[d][0] <= inf[d]) if mask[d][1] else (mask[d][0] >= sup[d]) for d in range(dim)])]
    if not new_mw:
        area = np.prod([high - low for (low, high) in zip(inf, sup)])
        histogram[len(cover_alls)] += area
    else:
        mask, _ = random.choice(new_mw)
        d_lim = random.choice([d for d in range(dim) if inf[d] < mask[d][0] < sup[d]])
        lim = mask[d_lim][0]
        winners_distribution(inf, [lim if d == d_lim else sup[d] for d in range(dim)], new_mw, histogram, cover_alls)
        winners_distribution([lim if d == d_lim else inf[d] for d in range(dim)], sup, new_mw, histogram, cover_alls)
    return histogram


def random_mask(dim):
    """Random mask.

    Parameters
    ----------
    dim : int
        Dimension of the Euclidean space under study.

    Returns
    -------
    list of tuple
        A mask of dimension `dim`, whose limits are in [0, 1]. For the definition of a mask, cf. :meth:`masks_area`.
    """
    return [(random.random(), random.choice([True, False])) for _ in range(dim)]


def random_masks(dim, n_masks):
    """List of random masks.

    Parameters
    ----------
    dim : int
        Dimension of the Euclidean space under study.
    n_masks : int
        Number of masks.

    Returns
    -------
    list of list of tuple
        A list of `n_masks` masks of dimension `d`, whose limits are in [0, 1]. For the definition of a mask, cf.
        :meth:`masks_area`.

    Examples
    --------
        >>> dim = 6
        >>> n_masks = 4
        >>> masks = random_masks(dim=dim, n_masks=n_masks)
        >>> isclose(masks_area(inf=[0]*dim, sup=[1]*dim, masks=masks),
        ...         masks_area_naive(inf=[0]*dim, sup=[1]*dim, masks=masks))
        True
    """
    return [random_mask(dim) for _ in range(n_masks)]



from random import randint
from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex
from poisson_approval.profiles.ProfileHistogram import ProfileHistogram


class GeneratorProfileHistogramSinglePeakedUniform:
    """A generator of single-peaked histogram-profiles (:class:`ProfileHistogram`) following the uniform distribution.

    Parameters
    ----------
    n_bins : int
        The numbers of bins for each histogram.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileHistogram`.

    Notes
    -----
    A candidate at random is chosen to be in the middle of the political axis. Then the ordinal profile is drawn
    uniformly on the simplex of the four possible rankings. Then each histogram is drawn uniformly on the simplex.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> generator = GeneratorProfileHistogramSinglePeakedUniform(n_bins=3)
        >>> profile = generator()
        >>> print(profile)
        <abc: 0.5488135039273248 [0.4236548  0.12122838 0.45511682], \
bac: 0.05394987214431912 [0.43758721 0.2083069  0.35410589], \
bca: 0.1124259903007756 [0.891773   0.07188976 0.03633724], \
cba: 0.2848106336275805 [0.38344152 0.40828352 0.20827496]> (Condorcet winner: a)
    """

    def __init__(self, n_bins, **kwargs):
        self.n_bins = n_bins
        self.kwargs = kwargs

    def __call__(self):
        """
        Returns
        -------
        ProfileHistogram
            A profile.
        """
        d_i_rankings = {
            0: ['abc', 'acb', 'bac', 'cab'],
            1: ['abc', 'bac', 'bca', 'cba'],
            2: ['acb', 'bca', 'cab', 'cba']
        }
        rankings = d_i_rankings[randint(0, 2)]
        x = rand_simplex(d=4)
        return ProfileHistogram({ranking: x[i] for i, ranking in enumerate(rankings)},
                                {ranking: rand_simplex(d=self.n_bins) for ranking in rankings},
                                **self.kwargs)



from poisson_approval import StrategyTwelve, PLURALITY, ANTI_PLURALITY, UTILITY_DEPENDENT


def test_strategy_twelve_plurality():
    """
        >>> strategy = StrategyTwelve({'abc': UTILITY_DEPENDENT, 'cba': 'b'}, voting_rule=PLURALITY)
        >>> strategy
        StrategyTwelve({'abc': 'utility-dependent', 'cba': 'b'}, voting_rule='Plurality')
        >>> print(strategy)
        <abc: utility-dependent, cba: b> (Plurality)
        >>> strategy.a_bc
        'a'
        >>> strategy.ab_c
        'b'
    """
    pass


def test_strategy_twelve_anti_plurality():
    """
        >>> strategy = StrategyTwelve({'abc': UTILITY_DEPENDENT, 'cba': 'bc'}, voting_rule=ANTI_PLURALITY)
        >>> strategy
        StrategyTwelve({'abc': 'utility-dependent', 'cba': 'bc'}, voting_rule='Anti-plurality')
        >>> print(strategy)
        <abc: utility-dependent, cba: bc> (Anti-plurality)
        >>> strategy.a_bc
        'ac'
        >>> strategy.ab_c
        'ab'
    """
    pass



def _cache(f):
    """Auxiliary decorator used by :meth:`cached_property`.

    Parameters
    ----------
    f : callable
        A method with no argument (except ``self``).

    Returns
    -------
    callable
        The same function, but with a `caching` behavior.

    Notes
    -----
    In practice, the values are stored in an attribute of the object that is a dictionary called ``_cached_properties``.
    For example, consider a method ``foo(self)``.

        * If the value is not computed yet, the decorated method will compute the value, store it in
          ``self._cached_properties['foo']`` and return it.
        * If the value is already computed, the decorated method will get it from ``self._cached_properties['foo']``
          and return it.
    """
    name = f.__name__

    # noinspection PyProtectedMember
    def _f(*args):
        try:
            return args[0]._cached_properties[name]
        except (KeyError, AttributeError):
            value = f(*args)
            try:
                # Not stored in cache
                args[0]._cached_properties[name] = value
            except AttributeError:
                # Cache does not even exist
                args[0]._cached_properties = {name: value}
            return value
    _f.__doc__ = f.__doc__
    return _f


def cached_property(f):
    """Decorator used in replacement of ``@property`` to put the value in cache automatically.

    Notes
    -----
    The first time the attribute is used, it is computed on-demand and put in cache. Later accesses to the
    attributes will use the cached value.

    Cf. :class:`DeleteCacheMixin` for an example.
    """
    return property(_cache(f))


class DeleteCacheMixin:
    """Mixin used to delete cached properties.

    Notes
    -----
    Cf. decorator :meth:`cached_property`.

    Examples
    --------
        >>> class Example(DeleteCacheMixin):
        ...     @cached_property
        ...     def x(self):
        ...         print('Big computation...')
        ...         return 6 * 7
        >>> a = Example()
        >>> a.x
        Big computation...
        42
        >>> a.x
        42
        >>> a.delete_cache()
        >>> a.x
        Big computation...
        42
    """

    # noinspection PyAttributeOutsideInit
    def delete_cache(self) -> None:
        self._cached_properties = dict()


def property_deleting_cache(hidden_variable_name, doc=''):
    """Define a property that deletes the cache when set or deleted.

    Parameters
    ----------
    hidden_variable_name : str
        The name of the hidden variable used to store the value of the property.
    doc : str
        The docstring of the property.

    Notes
    -----
    The class must inherit from :class:`DeleteCacheMixin`.

    Examples
    --------

    >>> class MyClass(DeleteCacheMixin):
    ...     def __init__(self, some_parameter):
    ...         self.some_parameter = some_parameter
    ...     some_parameter = property_deleting_cache('_some_parameter')
    ...     @cached_property
    ...     def my_cached_property(self):
    ...         print('Computing my_cached_property...')
    ...         return 'Hello %s!' % self.some_parameter
    >>> my_object = MyClass(some_parameter='World')
    >>> my_object.my_cached_property
    Computing my_cached_property...
    'Hello World!'
    >>> my_object.my_cached_property
    'Hello World!'
    >>> my_object.some_parameter = 'everybody'
    >>> my_object.my_cached_property
    Computing my_cached_property...
    'Hello everybody!'
    """

    def getter(self):
        return getattr(self, hidden_variable_name)

    def setter(self, value):
        self.delete_cache()
        setattr(self, hidden_variable_name, value)

    def deleter(self):
        self.delete_cache()
        delattr(self, hidden_variable_name)

    return property(getter, setter, deleter, doc)



import math
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.constants.Focus import Focus
from fractions import Fraction


class ExploreGridTaus:
    """Explore a grid of tau-vectors.

    Parameters
    ----------
    denominator : int or list of int
        The grain of the grid. For example, if ``denominator=9``, we examine tau-vectors such as
        (1/9, 2/9, 2/9, 1/9, 1/3, 0) or (1/9, 5/9, 0, 3/9, 0, 0), etc. If a list of integers is given, then all the
        corresponding denominators are examined.
    test : callable
        A function ``TauVector -> bool``. Only the tau-vectors satisfying this test will be examined.
        Default: always True.
    standardized : bool
        If True, then we examine only the standardized tau-vectors (i.e. up to relabelling the candidates).
        Cf. :attr:`TauVector.is_standardized`.

    Examples
    --------
        >>> def test(tau):
        ...     return tau.focus == Focus.DIRECT
        >>> exploration = ExploreGridTaus(denominator=6, test=test)
        >>> exploration
        <a: 1/3, ab: 1/6, bc: 1/2> ==> b
        <a: 1/3, b: 1/6, bc: 1/2> ==> b
        <a: 1/3, ab: 1/6, ac: 1/3, b: 1/6> ==> a
        <a: 1/3, ac: 1/6, b: 1/3, c: 1/6> ==> a
        <a: 1/2, ab: 1/3, ac: 1/6> ==> a
        <a: 1/2, ac: 1/3, b: 1/6> ==> a
        <a: 1/2, ab: 1/6, b: 1/6, c: 1/6> ==> a
        >>> print(exploration[0])
        <a: 1/3, ab: 1/6, bc: 1/2> ==> b
    """

    def __init__(self, denominator, test=None, standardized=True):
        if type(denominator) == int:
            denominators = [denominator]
        else:
            denominators = denominator
        if test is None:
            def test(_):
                return True
        self.taus = []
        for d in denominators:
            tau_1_min = int(math.ceil(d / 6)) if standardized else 0
            for tau_1 in range(tau_1_min, d + 1):
                tau_2_max = d - tau_1
                if standardized:
                    tau_2_max = min(tau_1, tau_2_max)
                for tau_2 in range(0, tau_2_max + 1):
                    tau_3_max = d - tau_1 - tau_2
                    if standardized:
                        tau_3_max = min(tau_1, tau_3_max)
                    for tau_3 in range(0, tau_3_max + 1):
                        tau_4_max = d - tau_1 - tau_2 - tau_3
                        if standardized:
                            tau_4_max = min(tau_1, tau_4_max)
                        for tau_4 in range(0, tau_4_max + 1):
                            tau_5_max = d - tau_1 - tau_2 - tau_3 - tau_4
                            if standardized:
                                tau_5_max = min(tau_1, tau_5_max)
                            for tau_5 in range(0, tau_5_max + 1):
                                tau_6 = (
                                    d - tau_1 - tau_2 - tau_3 - tau_4 - tau_5)
                                tau = TauVector({
                                    'a': Fraction(tau_1, d),
                                    'b': Fraction(tau_2, d),
                                    'c': Fraction(tau_3, d),
                                    'ab': Fraction(tau_4, d),
                                    'ac': Fraction(tau_5, d),
                                    'bc': Fraction(tau_6, d)
                                })
                                if standardized and not tau.is_standardized:
                                    continue
                                if not test(tau):
                                    continue
                                self.taus.append(tau)

    def __getitem__(self, item):
        return self.taus[item]

    def __repr__(self):
        return '\n'.join([str(tau) for tau in self.taus])



import pytest
from poisson_approval.utils.UtilCache import cached_property, DeleteCacheMixin, property_deleting_cache


def test():
    """
    In the following example, in the last instruction, foo must not be computed again:

        >>> class MyClass:
        ...
        ...     @cached_property
        ...     def foo(self):
        ...         print('Computing foo...')
        ...         return 'result_foo'
        ...
        ...     @cached_property
        ...     def foobar(self):
        ...         print('Computing foobar...')
        ...         return self.foo + 'bar'
        >>> my_object = MyClass()
        >>> print(my_object.foobar)
        Computing foobar...
        Computing foo...
        result_foobar
        >>> print(my_object.foo)
        result_foo
    """
    pass


def test_property_deleting_cache():

    class MyClass(DeleteCacheMixin):
        def __init__(self, some_parameter):
            self.some_parameter = some_parameter
        some_parameter = property_deleting_cache('_some_parameter')

        @cached_property
        def my_cached_property(self):
            print('Computing my_cached_property...')
            return 'Hello %s!' % self.some_parameter

    my_object = MyClass(some_parameter='World')
    assert my_object.my_cached_property == 'Hello World!'
    del my_object.some_parameter
    with pytest.raises(AttributeError):
        print(my_object.my_cached_property)



import warnings
from math import isclose
from poisson_approval.constants.constants import *
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.containers.AnalyzedStrategies import AnalyzedStrategies
from poisson_approval.profiles.ProfileCardinal import ProfileCardinal
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.Util import product_dict
from poisson_approval.utils.UtilCache import cached_property


class ProfileDiscrete(ProfileCardinal):
    """Profile with a discrete distribution of voters.

    Parameters
    ----------
    d : dict
        The first possible format is a dict of dict that, to a ranking (first key) and a utility (second key),
        associates the share of voters who have this ranking and this utility for their second candidate. The second
        possible format is a dict that, to a tuple (ranking, utility), associates the corresponding share of voters.
        The two formats can be mixed in the same profile declaration (cf. example below).
    normalization_warning : bool
        Whether a warning should be issued if the input distribution is not normalized.
    ratio_sincere : Number
        The ratio of sincere voters, in the interval [0, 1]. This is used for :meth:`tau`.
    ratio_fanatic : Number
        The ratio of fanatic voters, in the interval [0, 1]. This is used for :meth:`tau`. The sum of `ratio_sincere`
        and `ratio_fanatic` must not exceed 1.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.

    Attributes
    ----------
    d_ranking_utility_share : dict of dict
        To a ranking (first key) and a utility (second key), itassociates the share of voters who have this ranking and
        this utility for their second candidate.

    Notes
    -----
    If the input distribution is not normalized, the profile will be normalized anyway and a warning is
    issued (unless `normalization_warning` is False).

    Examples
    --------
        >>> from fractions import Fraction
        >>> profile = ProfileDiscrete({
        ...     'abc': {0.3: Fraction(26, 100), 0.8: Fraction(53, 100)},
        ...     ('bac', 0.1): Fraction(21, 100)
        ... })
        >>> profile
        ProfileDiscrete({'abc': {0.3: Fraction(13, 50), 0.8: Fraction(53, 100)}, 'bac': {0.1: Fraction(21, 100)}})
        >>> print(profile)
        <abc 0.3: 13/50, abc 0.8: 53/100, bac 0.1: 21/100> (Condorcet winner: a)
        >>> profile.d_ranking_share
        {'abc': Fraction(79, 100), 'bac': Fraction(21, 100)}
        >>> profile.abc
        Fraction(79, 100)
        >>> profile.have_ranking_with_utility_above_u('abc', 0.5)
        Fraction(53, 100)
        >>> profile.have_ranking_with_utility_below_u('abc', 0.5)
        Fraction(13, 50)
        >>> profile.have_ranking_with_utility_u('abc', 0.3)
        Fraction(13, 50)
    """

    def __init__(self, d, normalization_warning=True, ratio_sincere=0, ratio_fanatic=0, voting_rule=APPROVAL):
        """
            >>> profile = ProfileDiscrete({42: 51})
            Traceback (most recent call last):
            TypeError: Key should be tuple or str, got: <class 'int'> instead.
        """
        super().__init__(ratio_sincere=ratio_sincere, ratio_fanatic=ratio_fanatic, voting_rule=voting_rule)
        self.d_ranking_utility_share = DictPrintingInOrderIgnoringZeros({
            ranking: DictPrintingInOrderIgnoringZeros() for ranking in RANKINGS})
        for key, value in d.items():
            if isinstance(key, tuple):
                ranking, utility = key
                share = value
                if share > 0:
                    self.d_ranking_utility_share[ranking][utility] = (
                        self.d_ranking_utility_share[ranking].get(utility, 0) + share)
            elif isinstance(key, str):
                ranking = key
                d_utility_share = value
                for utility, share in d_utility_share.items():
                    if share > 0:
                        self.d_ranking_utility_share[ranking][utility] = (
                            self.d_ranking_utility_share[ranking].get(utility, 0) + share)
            else:
                raise TypeError('Key should be tuple or str, got: %s instead.' % type(key))
        # Normalize if necessary
        total = sum([sum(d_utility_share.values()) for d_utility_share in self.d_ranking_utility_share.values()])
        if not isclose(total, 1.):
            if normalization_warning:
                warnings.warn("Warning: profile is not normalized, I will normalize it.")
            for d_utility_share in self.d_ranking_utility_share.values():
                for utility, share in d_utility_share.items():
                    d_utility_share[utility] = share / total

    @cached_property
    def d_ranking_share(self):
        return DictPrintingInOrderIgnoringZeros({
            ranking: sum(d_utility_share.values()) for ranking, d_utility_share in self.d_ranking_utility_share.items()
        })

    def have_ranking_with_utility_above_u(self, ranking, u):
        d_utility_share = self.d_ranking_utility_share[ranking]
        return sum([share for utility, share in d_utility_share.items() if utility > u])

    def have_ranking_with_utility_u(self, ranking, u):
        d_utility_share = self.d_ranking_utility_share[ranking]
        return sum([share for utility, share in d_utility_share.items() if utility == u])

    def have_ranking_with_utility_below_u(self, ranking, u):
        d_utility_share = self.d_ranking_utility_share[ranking]
        return sum([share for utility, share in d_utility_share.items() if utility < u])

    def __repr__(self):
        """
        >>> from fractions import Fraction
        >>> profile = ProfileDiscrete({
        ...     'abc': {0.3: Fraction(26, 100), 0.8: Fraction(53, 100)},
        ...     ('bac', 0.1): Fraction(21, 100)
        ... }, ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(2, 10))
        >>> profile
        ProfileDiscrete({'abc': {0.3: Fraction(13, 50), 0.8: Fraction(53, 100)}, 'bac': {0.1: Fraction(21, 100)}}, \
ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
        """
        arguments = repr(self.d_ranking_utility_share)
        if self.ratio_sincere > 0:
            arguments += ', ratio_sincere=%r' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            arguments += ', ratio_fanatic=%r' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'ProfileDiscrete(%s)' % arguments

    def __str__(self):
        """
        >>> from fractions import Fraction
        >>> profile = ProfileDiscrete({
        ...     'abc': {0.3: Fraction(26, 100), 0.8: Fraction(53, 100)},
        ...     ('bac', 0.1): Fraction(21, 100)
        ... }, ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(2, 10))
        >>> print(profile)
        <abc 0.3: 13/50, abc 0.8: 53/100, bac 0.1: 21/100> (Condorcet winner: a) (ratio_sincere: 1/10) \
(ratio_fanatic: 1/5)
        """
        result = '<' + ', '.join([
            '%s %s: %s' % (ranking, utility, self.d_ranking_utility_share[ranking][utility])
            for ranking in sorted(self.d_ranking_utility_share.keys()) if self.d_ranking_utility_share[ranking]
            for utility in sorted(self.d_ranking_utility_share[ranking].keys())
        ]) + '>'
        if self.is_profile_condorcet:
            result += ' (Condorcet winner: %s)' % self.condorcet_winners
        if self.ratio_sincere > 0:
            result += ' (ratio_sincere: %s)' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            result += ' (ratio_fanatic: %s)' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            result += ' (%s)' % self.voting_rule
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : Object

        Returns
        -------
        bool
            True iff this profile is equal to `other`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileDiscrete({
            ...     'abc': {0.3: Fraction(26, 100), 0.8: Fraction(53, 100)},
            ...     ('bac', 0.1): Fraction(21, 100)
            ... })
            >>> profile == ProfileDiscrete({
            ...     'abc': {0.3: Fraction(26, 100), 0.8: Fraction(53, 100)},
            ...     ('bac', 0.1): Fraction(21, 100)
            ... })
            True
        """
        return (isinstance(other, ProfileDiscrete)
                and all([self.d_ranking_utility_share[ranking] == other.d_ranking_utility_share[ranking]
                         for ranking in RANKINGS])
                and self.ratio_sincere == other.ratio_sincere
                and self.ratio_fanatic == other.ratio_fanatic
                and self.voting_rule == other.voting_rule)

    @cached_property
    def standardized_version(self):
        """ProfileDiscrete : Standardized version of the profile (makes it unique, up to permutations of the candidates).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileDiscrete({
            ...     'abc': {0.3: Fraction(26, 100), 0.8: Fraction(53, 100)},
            ...     ('bac', 0.1): Fraction(21, 100)
            ... })
            >>> print(profile.standardized_version)
            <abc 0.3: 13/50, abc 0.8: 53/100, bac 0.1: 21/100> (Condorcet winner: a)
            >>> profile.is_standardized
            True
        """
        def translate(s, permute):
            return s.replace('a', permute[0]).replace('b', permute[1]).replace('c', permute[2])

        best_d = {}
        best_signature = []
        for perm in XYZ_PERMUTATIONS:
            d_test = {translate(ranking, perm): d_utility_share
                      for ranking, d_utility_share in self.d_ranking_utility_share.items()}
            signature_test = [[(utility, d_test[ranking][utility]) for utility in sorted(d_test[ranking].keys())]
                              for ranking in XYZ_RANKINGS]
            if signature_test > best_signature:
                best_signature = signature_test
                best_d = d_test
        return ProfileDiscrete({ranking: best_d[xyz_ranking] for ranking, xyz_ranking in zip(RANKINGS, XYZ_RANKINGS)},
                               ratio_sincere=self.ratio_sincere, ratio_fanatic=self.ratio_fanatic,
                               voting_rule=self.voting_rule)

    @cached_property
    def analyzed_strategies(self):
        """AnalyzedStrategies : Analyzed strategies of the profile.

        Examples
        --------
            >>> from fractions import Fraction
            >>> from fractions import Fraction
            >>> profile = ProfileDiscrete({
            ...     'abc': {0.3: Fraction(26, 100), 0.8: Fraction(53, 100)},
            ...     ('bac', 0.1): Fraction(21, 100)
            ... })
            >>> profile.analyzed_strategies
            Equilibrium:
            <abc: a, bac: b> ==> a (FF)
            <BLANKLINE>
            Non-equilibria:
            <abc: ab, bac: ab> ==> a, b (FF)
            <abc: ab, bac: b> ==> b (FF)
            <abc: utility-dependent (0.55), bac: ab> ==> a (FF)
            <abc: utility-dependent (0.55), bac: b> ==> a (FF)
            <abc: a, bac: ab> ==> a (FF)
            >>> print(profile.winners_at_equilibrium)
            a
        """
        equilibria = []
        utility_dependent = []
        inconclusive = []
        non_equilibria = []

        def possible_thresholds(ranking):
            if self.d_ranking_share[ranking] == 0:
                return [None]
            d_utility_share = self.d_ranking_utility_share[ranking]
            utilities = sorted(d_utility_share.keys())
            return [0] + [(x +y) / 2 for x, y in zip(utilities[:-1], utilities[1:])] + [1]

        d_ranking_possible_thresholds = {ranking: possible_thresholds(ranking) for ranking in RANKINGS}

        for d_ranking_threshold in product_dict(d_ranking_possible_thresholds):
            strategy = StrategyThreshold(d_ranking_threshold, profile=self)
            status = strategy.is_equilibrium
            if status == EquilibriumStatus.EQUILIBRIUM:
                equilibria.append(strategy)
            elif status == EquilibriumStatus.UTILITY_DEPENDENT:  # pragma: no cover
                utility_dependent.append(strategy)
                warnings.warn('Met a utility-dependent case: \nprofile = %r\nstrategy = %r' % (self, strategy))
            elif status == EquilibriumStatus.INCONCLUSIVE:  # pragma: no cover
                inconclusive.append(strategy)
                warnings.warn('Met an inconclusive case: \nprofile = %r\nstrategy = %r' % (self, strategy))
            else:
                non_equilibria.append(strategy)
        return AnalyzedStrategies(equilibria, utility_dependent, inconclusive, non_equilibria)



from poisson_approval import ProfileTwelve, StrategyTwelve


def test_deepcopy_with_attached_profile():
    """
    >>> strategy = StrategyTwelve(d_ranking_ballot={'abc': 'ab'})
    >>> print(strategy)
    <abc: ab>
    >>> profile = ProfileTwelve(d_type_share={'a_bc': 1, 'ab_c': 1})
    >>> strategy_with_profile = strategy.deepcopy_with_attached_profile(profile)
    >>> print(strategy_with_profile)
    <abc: ab> ==> a, b
    """
    pass



from poisson_approval import ExploreGridProfilesOrdinal


def test():
    """
        >>> exploration = ExploreGridProfilesOrdinal(denominator=[3, 4])
        >>> exploration
        0 equilibrium, 1 utility-dependent equilibrium, 3 non-equilibria
        <abc: 2/3, cab: 1/3> (Condorcet winner: a)
        <abc: 1/2, cba: 1/2> (Condorcet winner: a, b, c)
        <abc: 1/2, bca: 1/2> (Condorcet winner: a, b)
        <abc: 3/4, cab: 1/4> (Condorcet winner: a)
        <BLANKLINE>
        0 equilibrium, 1 utility-dependent equilibrium, 7 non-equilibria
        <abc: 1/3, acb: 1/3, bac: 1/3> (Condorcet winner: a)
        <abc: 1/2, bca: 1/4, cab: 1/4> (Condorcet winner: a)
        <abc: 1/2, acb: 1/4, cab: 1/4> (Condorcet winner: a)
        <abc: 1/2, acb: 1/4, bac: 1/4> (Condorcet winner: a)
        <BLANKLINE>
        0 equilibrium, 2 utility-dependent equilibria, 2 non-equilibria
        <abc: 2/3, cba: 1/3> (Condorcet winner: a)
        <abc: 2/3, bca: 1/3> (Condorcet winner: a)
        <abc: 3/4, cba: 1/4> (Condorcet winner: a)
        <abc: 3/4, bca: 1/4> (Condorcet winner: a)
        <BLANKLINE>
        0 equilibrium, 2 utility-dependent equilibria, 6 non-equilibria
        <abc: 1/3, bca: 1/3, cab: 1/3>
        <abc: 1/2, cab: 1/4, cba: 1/4> (Condorcet winner: a, c)
        <abc: 1/2, bac: 1/4, bca: 1/4> (Condorcet winner: a, b)
        <abc: 1/2, acb: 1/4, cba: 1/4> (Condorcet winner: a)
        <abc: 1/2, acb: 1/4, bca: 1/4> (Condorcet winner: a)
        <BLANKLINE>
        0 equilibrium, 3 utility-dependent equilibria, 5 non-equilibria
        <abc: 1/3, acb: 1/3, bca: 1/3> (Condorcet winner: a)
        <BLANKLINE>
        0 equilibrium, 4 utility-dependent equilibria, 12 non-equilibria
        <abc: 1/4, acb: 1/4, bac: 1/4, cba: 1/4> (Condorcet winner: a, b)
        <abc: 1/4, acb: 1/4, bac: 1/4, bca: 1/4> (Condorcet winner: a, b)
        <BLANKLINE>
        0 equilibrium, 7 utility-dependent equilibria, 9 non-equilibria
        <abc: 1/4, acb: 1/4, bca: 1/4, cba: 1/4> (Condorcet winner: a, b, c)
        <BLANKLINE>
        1 equilibrium, 0 utility-dependent equilibrium, 1 non-equilibrium
        <abc: 1> (Condorcet winner: a)
        <BLANKLINE>
        1 equilibrium, 0 utility-dependent equilibrium, 3 non-equilibria
        <abc: 2/3, bac: 1/3> (Condorcet winner: a)
        <abc: 2/3, acb: 1/3> (Condorcet winner: a)
        <abc: 1/2, bac: 1/2> (Condorcet winner: a, b)
        <abc: 1/2, acb: 1/2> (Condorcet winner: a)
        <abc: 3/4, bac: 1/4> (Condorcet winner: a)
        <abc: 3/4, acb: 1/4> (Condorcet winner: a)
        <BLANKLINE>
        1 equilibrium, 2 utility-dependent equilibria, 5 non-equilibria
        <abc: 1/2, bca: 1/4, cba: 1/4> (Condorcet winner: a, b)
        <BLANKLINE>
        2 equilibria, 0 utility-dependent equilibrium, 6 non-equilibria
        <abc: 1/2, bac: 1/4, cba: 1/4> (Condorcet winner: a, b)
        <BLANKLINE>
        2 equilibria, 1 utility-dependent equilibrium, 5 non-equilibria
        <abc: 1/3, bac: 1/3, cab: 1/3> (Condorcet winner: a)
        <abc: 1/2, bac: 1/4, cab: 1/4> (Condorcet winner: a)
        <BLANKLINE>
        2 equilibria, 2 utility-dependent equilibria, 12 non-equilibria
        <abc: 1/4, acb: 1/4, bac: 1/4, cab: 1/4> (Condorcet winner: a)
        <BLANKLINE>
        >>> len(exploration.items())
        13
        >>> len(exploration.keys())
        13
        >>> len(exploration.values())
        13
    """
    pass



from poisson_approval import TauVector, EventTrio2t


def test():
    """
        >>> tau = TauVector({'a': 2, 'b': 3, 'c': 5, 'ab': 7, 'ac': 11, 'bc':13}, normalization_warning=False)
        >>> tau.trio_2t_ab
        <asymptotic = exp(- 0.0392799 n + ? log n + ? + o(1)), phi_a = 1.35481, phi_b = 1.08479, phi_c = 0.680414, \
phi_ab = 1.46969, phi_ac = 0.921834, phi_bc = 0.738109>
    """
    pass



from copy import deepcopy
from poisson_approval.constants.constants import *
from poisson_approval.utils.UtilCache import cached_property, DeleteCacheMixin, property_deleting_cache


class Strategy(DeleteCacheMixin):
    """A strategy profile (abstract class).

    Parameters
    ----------
    profile : Profile, optional
        The "context" in which the strategy is used.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.
    """

    def __init__(self, profile=None, voting_rule=None):
        self.profile = profile  # Store the profile (if any)
        self.voting_rule = self._get_voting_rule_(profile, voting_rule)

    profile = property_deleting_cache('_profile')
    voting_rule = property_deleting_cache('_voting_rule')

    @staticmethod
    def _get_voting_rule_(profile, voting_rule):
        if voting_rule is None:
            if profile is not None:
                return profile.voting_rule
            else:
                return APPROVAL
        return voting_rule

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    def deepcopy_with_attached_profile(self, profile):
        """Deep copy with an attached profile.

        Parameters
        ----------
        profile : Profile

        Returns
        -------
        Strategy
            A deep copy of this strategy, with the attached profile `profile`.
        """
        strategy = deepcopy(self)
        strategy.profile = profile
        return strategy

    # Additional stuff when a profile is given
    # For tests of this, see file ``test_strategy_embed_profile.py``.

    @cached_property
    def is_equilibrium(self):
        """EquilibriumStatus : Whether this strategy is an equilibrium (in the context of the given profile). Cf.
        :meth:`Profile.is_equilibrium`.
        """
        if self.profile is not None:
            return self.profile.is_equilibrium(self)

    @cached_property
    def tau(self):
        """TauVector : The tau-vector associated to this strategy (in the context of the given profile). Cf.
        :meth:`Profile.tau`.
        """
        if self.profile is not None:
            return self.profile.tau(self)

    # noinspection NonAsciiCharacters
    @property
    def τ(self):
        """TauVector : The tau-vector (alternate notation). Cf. :meth:`Profile.τ`.
        """
        return self.tau


def make_method(name):
    def _method(self):
        if self.tau is not None:
            return getattr(self.tau, name)()
    _method.__doc__ = "Defined when a profile is given. Cf. :meth:`TauVector.%s`." % name
    return _method


for my_method in ['print_weak_pivots', 'print_all_pivots']:
    setattr(Strategy, my_method, make_method(my_method))


def make_property(name, doc):
    def _property(self):
        if self.tau is not None:
            return getattr(self.tau, name)
    _property.__doc__ = "Defined when a profile is given. Cf. :attr:`TauVector.%s`." % name
    return property(_property)


for my_property, my_doc in [('scores', 'The scores.'),
                            ('winners', 'The winners.'),
                            ('trio', 'Event: trio.'),
                            ('d_ranking_best_response', 'Best response profile.')]:
    setattr(Strategy, my_property, make_property(my_property, my_doc))

for my_property, my_doc in [
        ('trio_1t_', 'Event: trio of type 1t (this candidate has one point less than the other two.')]:
    for candidate in CANDIDATES:
        my_property_full = my_property + candidate
        setattr(Strategy, my_property_full, make_property(my_property_full, my_doc))

for my_property, my_doc in [
        ('duo_', 'Event: duo.'),
        ('pivot_weak_', 'Event: weak pivot.'),
        ('pivot_strict_', 'Event: strict pivot.'),
        ('trio_2t_', 'Event: trio of type 2t (these two candidates have one point less than the other one).')]:
    for pair in PAIRS_WITH_INVERSIONS:
        my_property_full = my_property + pair
        setattr(Strategy, my_property_full, make_property(my_property_full, my_doc))

for my_property, my_doc in [('pivot_tij_', 'Event: `personalized pivot` of type Tij.'),
                            ('pivot_tjk_', 'Event: `personalized pivot` of type Tjk.')]:
    for ranking in RANKINGS:
        my_property_full = my_property + ranking
        setattr(Strategy, my_property_full, make_property(my_property_full, my_doc))



from fractions import Fraction
from poisson_approval.best_response.BestResponse import BestResponse
from poisson_approval.constants.constants import *
from poisson_approval.utils.UtilCache import cached_property


class BestResponseAntiPlurality(BestResponse):
    """Best response for a given ordinal type of voter in Anti-plurality.

    The main objective of this class is to compute :attr:`threshold_utility`.

    For the sake of consistency with :class:`BestResponseApproval`, it provides the string :attr:`justification`,
    indicating which sub-algorithm was used. But since there are no actual sub-algorithms for anti-plurality, the
    justification is always the same: 'Anti-plurality analysis'.

    Parameters
    ----------
    Cf. :class:`BestResponse`.

    Attributes
    ----------
    Cf. :class:`BestResponse`
    """

    ANTI_PLURALITY_ANALYSIS = 'Anti-plurality analysis'
    voting_rule = ANTI_PLURALITY

    @cached_property
    def results(self):
        assert self.tau.voting_rule == ANTI_PLURALITY
        tau_minus_i = self.tau_jk
        tau_minus_j = self.tau_ik
        tau_minus_k = self.tau_ij
        if tau_minus_k > tau_minus_i and tau_minus_k > tau_minus_j:
            # The best response is `- j`, i.e. `ik`
            threshold_utility = 1
        elif tau_minus_i == tau_minus_k > tau_minus_j:
            threshold_utility = Fraction(1, 2)
        else:
            # The best response is `- k`, i.e. `ij`
            threshold_utility = 0
        return threshold_utility, self.ANTI_PLURALITY_ANALYSIS



import numpy as np
from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold


class GeneratorStrategyThresholdUniform:
    """A generator of threshold strategies (:class:`StrategyThreshold`) following the uniform distribution.

    Parameters
    ----------
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`StrategyThreshold`.

    Notes
    -----
    Each threshold is drawn uniformly in the interval [0, 1].

    Examples
    --------
        >>> initialize_random_seeds()
        >>> generator = GeneratorStrategyThresholdUniform()
        >>> strategy = generator()
        >>> print(strategy)
        <abc: utility-dependent (0.5488135039273248), acb: utility-dependent (0.7151893663724195), \
bac: utility-dependent (0.6027633760716439), bca: utility-dependent (0.5448831829968969), \
cab: utility-dependent (0.4236547993389047), cba: utility-dependent (0.6458941130666561)>
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self):
        """
        Returns
        -------
        StrategyThreshold
            A strategy.
        """
        return StrategyThreshold({ranking: np.random.rand() for ranking in RANKINGS}, **self.kwargs)



RANKINGS = ['abc', 'acb', 'bac', 'bca', 'cab', 'cba']
PERMUTATIONS = RANKINGS
TWELVE_TYPES = ['a_bc', 'ab_c', 'a_cb', 'ac_b', 'b_ac', 'ba_c', 'b_ca', 'bc_a', 'c_ab', 'ca_b', 'c_ba', 'cb_a']

CANDIDATES = ['a', 'b', 'c']
PAIRS_WITHOUT_INVERSIONS = ['ab', 'ac', 'bc']
PAIRS_INVERTED = ['ba', 'ca', 'cb']
PAIRS_WITH_INVERSIONS = PAIRS_WITHOUT_INVERSIONS + PAIRS_INVERTED
BALLOTS_WITHOUT_INVERSIONS = CANDIDATES + PAIRS_WITHOUT_INVERSIONS
BALLOTS_WITH_INVERSIONS = BALLOTS_WITHOUT_INVERSIONS + PAIRS_INVERTED

UTILITY_DEPENDENT = 'utility-dependent'
INCONCLUSIVE = 'inconclusive'

APPROVAL = 'Approval'
PLURALITY = 'Plurality'
ANTI_PLURALITY = 'Anti-plurality'


def _f_abc_xyz(my_list):
    return [element.replace('a', 'x').replace('b', 'y').replace('c', 'z') for element in my_list]


XYZ_RANKINGS = _f_abc_xyz(RANKINGS)
XYZ_PERMUTATIONS = XYZ_RANKINGS
XYZ_TWELVE_TYPES = _f_abc_xyz(TWELVE_TYPES)
XYZ_BALLOTS_WITHOUT_INVERSION = _f_abc_xyz(BALLOTS_WITHOUT_INVERSIONS)
XYZ_BALLOTS_INVERTED = _f_abc_xyz(PAIRS_INVERTED)
XYZ_BALLOTS_WITH_INVERSIONS = _f_abc_xyz(BALLOTS_WITH_INVERSIONS)



from fractions import Fraction
from poisson_approval.best_response.BestResponse import BestResponse
from poisson_approval.constants.constants import *
from poisson_approval.utils.UtilCache import cached_property


class BestResponsePlurality(BestResponse):
    """Best response for a given ordinal type of voter in Plurality.

    The main objective of this class is to compute :attr:`threshold_utility`.

    For the sake of consistency with :class:`BestResponseApproval`, it provides the string :attr:`justification`,
    indicating which sub-algorithm was used. But since there are no actual sub-algorithms for plurality, the
    justification is always the same: 'Plurality analysis'.

    Parameters
    ----------
    Cf. :class:`BestResponse`.

    Attributes
    ----------
    Cf. :class:`BestResponse`
    """

    PLURALITY_ANALYSIS = 'Plurality analysis'
    voting_rule = PLURALITY

    @cached_property
    def results(self):
        assert self.tau.voting_rule == PLURALITY
        if self.tau_i < self.tau_j and self.tau_i < self.tau_k:
            # The best response is `j`.
            threshold_utility = 0
        elif 0 < self.tau_i == self.tau_k < self.tau_j:
            threshold_utility = Fraction(1, 2)
        else:
            # The best response is `i`.
            threshold_utility = 1
        return threshold_utility, self.PLURALITY_ANALYSIS



from poisson_approval.constants.constants import *
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.Util import ballot_one, ballot_two, ballot_one_two, ballot_one_three


class StrategyOrdinal(StrategyThreshold):
    """A strategy profile for an ordinal preference profile.

    Parameters
    ----------
    d_ranking_ballot : dict
        Keys are rankings and values are ballots, e.g. ``'abc': 'ab'``. A
        ballot can be ``''`` if the behavior of these voters is not specified in the strategy.
    profile : Profile, optional
        The "context" in which the strategy is used.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.

    Examples
    --------
        >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
        >>> strategy
        StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
        >>> print(strategy)
        <abc: a, bac: ab, cab: c>
        >>> strategy.abc
        'a'
        >>> strategy.a_bc
        'a'
        >>> strategy.ab_c
        'a'
        >>> strategy.d_ranking_threshold['abc']
        1
    """

    def __init__(self, d_ranking_ballot, profile=None, voting_rule=None):
        """
            >>> strategy = StrategyOrdinal({'abc': 'non_existing_ballot'})
            Traceback (most recent call last):
            ValueError: Unknown strategy: non_existing_ballot
        """
        voting_rule = self._get_voting_rule_(profile, voting_rule)
        # Prepare the dictionary of thresholds
        d_ranking_threshold = DictPrintingInOrderIgnoringZeros()
        for ranking, ballot in d_ranking_ballot.items():
            if ballot == '':
                d_ranking_threshold[ranking] = None
            elif ballot == ballot_one(ranking) and voting_rule in {APPROVAL, PLURALITY}:
                d_ranking_threshold[ranking] = 1
            elif ballot == ballot_two(ranking) and voting_rule == PLURALITY:
                d_ranking_threshold[ranking] = 0
            elif (ballot in {ballot_one_two(ranking), ballot_one_two(ranking)[::-1]}
                  and voting_rule in {APPROVAL, ANTI_PLURALITY}):
                d_ranking_threshold[ranking] = 0
            elif (ballot in {ballot_one_three(ranking), ballot_one_three(ranking)[::-1]}
                  and voting_rule == ANTI_PLURALITY):
                d_ranking_threshold[ranking] = 1
            else:
                raise ValueError('Unknown strategy: ' + ballot)
        # Call parent class
        super().__init__(d_ranking_threshold=d_ranking_threshold, profile=profile, voting_rule=voting_rule)

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : object

        Returns
        -------
        bool
            True if this strategy is equal to `other`.

        Examples
        --------
            >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            >>> strategy == StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            True
        """
        return (isinstance(other, StrategyOrdinal)
                and self.d_ranking_ballot == other.d_ranking_ballot
                and self.voting_rule == other.voting_rule)

    def __repr__(self):
        arguments = repr(self.d_ranking_ballot)
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'StrategyOrdinal(%s)' % arguments

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')



from pytest import fixture
from fractions import Fraction
from poisson_approval import ProfileTwelve, StrategyTwelve, StrategyThreshold, EquilibriumStatus, PLURALITY, \
    ANTI_PLURALITY, UTILITY_DEPENDENT, TauVector


def test_iterative_voting_verbose():
    """
    >>> from fractions import Fraction
    >>> profile = ProfileTwelve(d_type_share={'a_bc': Fraction(1, 2), 'ab_c': Fraction(1, 2)})
    >>> strategy = StrategyTwelve(d_ranking_ballot={'abc': 'ab'})
    >>> result = profile.iterated_voting(init=strategy, n_max_episodes=100,
    ...                                  ballot_update_ratio=1, verbose=True)
    t = 0
    strategy: <abc: ab> ==> a, b
    tau_actual: <ab: 1> ==> a, b
    t = 1
    tau_perceived: <ab: 1> ==> a, b
    strategy: <abc: a> ==> a
    tau_full_response: <a: 1> ==> a
    tau_actual: <a: 1> ==> a
    t = 2
    tau_perceived: <a: 1> ==> a
    strategy: <abc: a> ==> a
    tau_full_response: <a: 1> ==> a
    tau_actual: <a: 1> ==> a
    t = 3
    tau_perceived: <a: 1> ==> a
    strategy: <abc: a> ==> a
    tau_full_response: <a: 1> ==> a
    tau_actual: <a: 1> ==> a
    >>> result = profile.fictitious_play(init=strategy, n_max_episodes=100,
    ...                                  perception_update_ratio=1, ballot_update_ratio=1, verbose=True)
    t = 0
    strategy: <abc: ab> ==> a, b
    tau_actual: <ab: 1> ==> a, b
    tau_perceived: <ab: 1> ==> a, b
    t = 1
    tau_perceived: <ab: 1> ==> a, b
    strategy: <abc: a> ==> a
    tau_full_response: <a: 1> ==> a
    tau_actual: <a: 1> ==> a
    t = 2
    tau_perceived: <a: 1> ==> a
    strategy: <abc: a> ==> a
    tau_full_response: <a: 1> ==> a
    tau_actual: <a: 1> ==> a
    """
    pass


@fixture()
def my_profile():
    return ProfileTwelve(d_type_share={'a_bc': 1, 'ab_c': 1})


@fixture()
def my_strategy():
    return StrategyTwelve(d_ranking_ballot={'abc': 'ab'})


def test_normalization(my_profile):
    assert my_profile.a_bc == 0.5


def test_not_equilibrium(my_profile, my_strategy):
    assert my_profile.is_equilibrium(my_strategy) == EquilibriumStatus.NOT_EQUILIBRIUM


def test_iterated_voting_without_convergence():
    """
        >>> my_profile = ProfileTwelve(d_type_share={'a_bc': 1, 'ab_c': 1})
        >>> my_strategy = StrategyTwelve(d_ranking_ballot={'abc': 'ab'})
        >>> result = my_profile.iterated_voting(init=my_strategy, n_max_episodes=1, ballot_update_ratio=1)
        >>> result['cycle_taus_actual']
        []
        >>> result['d_candidate_winning_frequency']
        {'a': Fraction(3, 4), 'b': Fraction(1, 4)}
    """
    pass


def test_fictitious_play_without_convergence(my_profile, my_strategy):
    """
        >>> my_profile = ProfileTwelve(d_type_share={'a_bc': 1, 'ab_c': 1})
        >>> my_strategy = StrategyTwelve(d_ranking_ballot={'abc': 'ab'})
        >>> result = my_profile.fictitious_play(init=my_strategy, n_max_episodes=1)
        >>> print(result['tau'])
        None
        >>> result['d_candidate_winning_frequency']
        {'a': Fraction(3, 4), 'b': Fraction(1, 4)}
    """
    pass


def test_plurality():
    """
        >>> profile = ProfileTwelve(
        ...     d_type_share={'a_bc': Fraction(1, 5), 'ba_c': Fraction(3, 5), 'c_ab': Fraction(1, 5)},
        ...     voting_rule=PLURALITY)
        >>> profile
        ProfileTwelve({'a_bc': Fraction(1, 5), 'ba_c': Fraction(3, 5), 'c_ab': Fraction(1, 5)}, voting_rule='Plurality')
        >>> print(profile)
        <a_bc: 1/5, ba_c: 3/5, c_ab: 1/5> (Condorcet winner: b) (Plurality)
        >>> profile.tau_sincere
        TauVector({'a': Fraction(1, 5), 'b': Fraction(3, 5), 'c': Fraction(1, 5)}, voting_rule='Plurality')
        >>> profile.tau_fanatic
        TauVector({'a': Fraction(1, 5), 'b': Fraction(3, 5), 'c': Fraction(1, 5)}, voting_rule='Plurality')
        >>> strategy = StrategyTwelve({'abc': 'a', 'bac': 'b', 'cab': 'c'}, profile=profile)
        >>> strategy.is_equilibrium
        EquilibriumStatus.EQUILIBRIUM
    """
    pass


def test_anti_plurality():
    """
        >>> profile = ProfileTwelve(
        ...     d_type_share={'ab_c': Fraction(2, 5), 'b_ca': Fraction(2, 5), 'ca_b': Fraction(1, 5)},
        ...     voting_rule=ANTI_PLURALITY)
        >>> profile
        ProfileTwelve({'ab_c': Fraction(2, 5), 'b_ca': Fraction(2, 5), 'ca_b': Fraction(1, 5)}, \
voting_rule='Anti-plurality')
        >>> print(profile)
        <ab_c: 2/5, b_ca: 2/5, ca_b: 1/5> (Anti-plurality)
        >>> profile.tau_sincere
        TauVector({'ab': Fraction(2, 5), 'ac': Fraction(1, 5), 'bc': Fraction(2, 5)}, voting_rule='Anti-plurality')
        >>> profile.tau_fanatic
        TauVector({'ab': Fraction(2, 5), 'ac': Fraction(1, 5), 'bc': Fraction(2, 5)}, voting_rule='Anti-plurality')
        >>> strategy = StrategyTwelve({'abc': 'ab', 'bca': 'bc', 'cab': 'ac'}, profile=profile)
        >>> strategy.is_equilibrium
        EquilibriumStatus.EQUILIBRIUM
    """
    pass


def test_initializer():
    """
        >>> profile = ProfileTwelve(
        ...     d_type_share={'ab_c': Fraction(2, 5), 'b_ca': Fraction(2, 5), 'ca_b': Fraction(1, 5)})
        >>> profile._initializer(init=StrategyTwelve({'abc': 'ab', 'bca': 'bc', 'cab': 'ac'}))
        (StrategyTwelve({'abc': 'ab', 'bca': 'bc', 'cab': 'ac'}), \
TauVector({'ab': Fraction(2, 5), 'ac': Fraction(1, 5), 'bc': Fraction(2, 5)}))
        >>> profile._initializer(TauVector({'ab': Fraction(2, 5), 'ac': Fraction(1, 5), 'bc': Fraction(2, 5)}))
        (None, TauVector({'ab': Fraction(2, 5), 'ac': Fraction(1, 5), 'bc': Fraction(2, 5)}))
        >>> profile._initializer('sincere')
        (None, TauVector({'ab': Fraction(2, 5), 'ac': Fraction(1, 5), 'b': Fraction(2, 5)}))
        >>> profile._initializer('fanatic')
        (None, TauVector({'a': Fraction(2, 5), 'b': Fraction(2, 5), 'c': Fraction(1, 5)}))
    """
    pass



import warnings
from math import isclose
from functools import partial
from fractions import Fraction
from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import sort_ballot
from poisson_approval.containers.Scores import Scores
from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.best_response.BestResponseApproval import BestResponseApproval
from poisson_approval.best_response.BestResponsePlurality import BestResponsePlurality
from poisson_approval.best_response.BestResponseAntiPlurality import BestResponseAntiPlurality
from poisson_approval.utils.UtilCache import cached_property
from poisson_approval.events.EventDuo import EventDuo
from poisson_approval.events.EventPivotStrict import EventPivotStrict
from poisson_approval.events.EventPivotTij import EventPivotTij
from poisson_approval.events.EventPivotTjk import EventPivotTjk
from poisson_approval.events.EventPivotWeak import EventPivotWeak
from poisson_approval.events.EventTrio import EventTrio
from poisson_approval.events.EventTrio1t import EventTrio1t
from poisson_approval.events.EventTrio2t import EventTrio2t
from poisson_approval.constants.Focus import Focus


# noinspection PyUnresolvedReferences
class TauVector:
    """A vector 'tau' (ballot distribution)

    Parameters
    ----------
    d_ballot_share : dict
        Ballot distribution, e.g. ``{'a': 0.1, 'ab': 0.6, 'c':0.3}``.
    normalization_warning : bool
        Whether a warning should be issued if the input distribution is not normalized.

    Notes
    -----
    If the input distribution `d_ballot_share` is not normalized, the tau vector will be normalized anyway and a
    warning is issued (unless `normalization_warning` is False).

    Examples
    --------
        >>> from fractions import Fraction
        >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
        >>> tau
        TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
        >>> print(tau)
        <a: 1/10, ab: 3/5, c: 3/10> ==> a
        >>> tau.a
        Fraction(1, 10)
        >>> tau.b
        0
        >>> tau.c
        Fraction(3, 10)
        >>> tau.ab
        Fraction(3, 5)
        >>> tau.ba  # Alternate notation for tau.ab
        Fraction(3, 5)
        >>> tau.ac
        0
        >>> tau.ca  # Alternate notation for tau.ac, etc.
        0
        >>> tau.bc
        0
        >>> tau.cb
        0
        >>> tau.duo_ab
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.duo_ba
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.duo_ac
        <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.duo_ca
        <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.duo_bc
        <asymptotic = exp(- 0.0514719 n - 0.5 log n - 0.836813 + o(1)), phi_a = 1, phi_c = 1.41421, phi_ab = 0.707107>
        >>> tau.duo_cb
        <asymptotic = exp(- 0.0514719 n - 0.5 log n - 0.836813 + o(1)), phi_a = 1, phi_c = 1.41421, phi_ab = 0.707107>
        >>> tau.pivot_weak_ab
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.pivot_weak_ba
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.pivot_weak_ac
        <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.pivot_weak_ca
        <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.pivot_weak_bc
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
        >>> tau.pivot_weak_cb
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = \
0.707107>
        >>> tau.pivot_strict_ab
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.pivot_strict_ba
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.pivot_strict_ac
        <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.pivot_strict_ca
        <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.pivot_strict_bc
        <asymptotic = exp(- inf)>
        >>> tau.pivot_strict_cb
        <asymptotic = exp(- inf)>
        >>> tau.pivot_tij_abc
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.pivot_tij_acb
        <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.371758 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.pivot_tij_bac
        <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.pivot_tij_bca
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.302013 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
        >>> tau.pivot_tij_cab
        <asymptotic = exp(- 0.0834849 n - 0.5 log n + 0.0518905 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.pivot_tij_cba
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
        >>> tau.pivot_tjk_abc
        <asymptotic = exp(- inf)>
        >>> tau.pivot_tjk_acb
        <asymptotic = exp(- inf)>
        >>> tau.pivot_tjk_bac
        <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.371758 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.pivot_tjk_bca
        <asymptotic = exp(- 0.0834849 n - 0.5 log n + 0.0518905 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
        >>> tau.pivot_tjk_cab
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.pivot_tjk_cba
        <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> tau.trio
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
        >>> tau.trio_1t_a
        <asymptotic = exp(- inf)>
        >>> tau.trio_1t_b
        <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.48597 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
        >>> tau.trio_1t_c
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.490239 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
        >>> tau.trio_2t_ab
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 1.18339 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
        >>> tau.trio_2t_ac
        <asymptotic = exp(- inf)>
        >>> tau.trio_2t_bc
        <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.1394 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
        >>> tau.trio_2t_ba
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 1.18339 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
        >>> tau.trio_2t_ca
        <asymptotic = exp(- inf)>
        >>> tau.trio_2t_cb
        <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.1394 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
    """

    def __init__(self, d_ballot_share: dict, normalization_warning: bool = True,
                 voting_rule=APPROVAL):
        """
            >>> tau = TauVector({'non_existing_ballot': 1})
            Traceback (most recent call last):
            ValueError: Unknown key: non_existing_ballot
        """
        # Populate the dictionary and check for typos in the input
        self.d_ballot_share = DictPrintingInOrderIgnoringZeros()
        for ballot, share in d_ballot_share.items():
            if ballot in BALLOTS_WITH_INVERSIONS:
                sorted_key = ''.join(sorted(ballot))
                self.d_ballot_share[sorted_key] = share
            else:
                raise ValueError('Unknown key: ' + ballot)
        for ballot in BALLOTS_WITHOUT_INVERSIONS:
            if ballot not in self.d_ballot_share:
                self.d_ballot_share[ballot] = 0
        # Normalize if necessary
        total = sum(self.d_ballot_share.values())
        if not isclose(total, 1., rel_tol=10E-4):
            if normalization_warning:
                warnings.warn("Warning: tau is not normalized, I will normalize it.")
            for ballot in self.d_ballot_share.keys():
                self.d_ballot_share[ballot] = self.d_ballot_share[ballot] / total
        # Voting rule
        self.voting_rule = voting_rule
        if self.voting_rule == PLURALITY:
            assert self.ab == self.ac == self.bc == 0
        elif self.voting_rule == ANTI_PLURALITY:
            assert self.a == self.b == self.c == 0

    def __repr__(self):
        arguments = repr(self.d_ballot_share)
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'TauVector(%s)' % arguments

    def __str__(self):
        s = '<%s>' % str(self.d_ballot_share)[1:-1] + ' ==> ' + str(self.winners)
        if self.voting_rule != APPROVAL:
            s += ' (%s)' % self.voting_rule
        return s

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    @cached_property
    def scores(self):
        """Scores : The scores.

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau.scores
            {'a': Fraction(7, 10), 'b': Fraction(3, 5), 'c': Fraction(3, 10)}
        """
        return Scores(dict(a=self.a + self.ab + self.ac,
                           b=self.b + self.ab + self.bc,
                           c=self.c + self.ac + self.bc))

    @property
    def winners(self):
        """Winners : The winners.

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau.winners
            Winners({'a'})
        """
        return self.scores.winners

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : Object

        Returns
        -------
        bool
            True iff it is the same tau-vector.

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau == TauVector({'a': Fraction(10, 100), 'ab': Fraction(60, 100), 'c': Fraction(30, 100)})
            True
        """
        return (isinstance(other, TauVector)
                and self.d_ballot_share == other.d_ballot_share
                and self.voting_rule == other.voting_rule)

    @cached_property
    def has_two_consecutive_zeros(self):
        """bool

        Whether the tau-vector has two consecutive holes in the "compass" representation. True iff
        ``self.a == 0 and self.ab == 0``, or ``self.ab == 0 and self.b == 0``, etc.
        """
        return ((self.a == 0 and (self.ab == 0 or self.ac == 0))
                or (self.b == 0 and (self.ab == 0 or self.bc == 0))
                or (self.c == 0 and (self.ac == 0 or self.bc == 0)))

    def isclose(self, other, *args, **kwargs):
        """Test near-equality.

        Parameters
        ----------
        other : Object
        *args
            Cf. :func:`math.isclose`.
        **kwargs
            Cf. :func:`math.isclose`.

        Returns
        -------
        isclose : bool
            True if this tau-vector is approximately equal to `other`.

        Examples
        --------
            >>> tau = TauVector({'ab': 0.4, 'b': 0.6})
            >>> tau.isclose(TauVector({'ab': 0.4, 'b': 0.59999999999999999999999999}))
            True
        """
        return isinstance(other, TauVector) and all([
            isclose(share, other.d_ballot_share[ballot], *args, **kwargs)
            for ballot, share in self.d_ballot_share.items()
        ])

    @cached_property
    def standardized_version(self):
        """Standardized version of the profile (makes it unique, up to permutations).

        Notes
        -----
        It returns the same profile, up to a permutation of the candidates. how the permutation is chosen in
        practice does not really matter: the important point is that the `standardized version` is the same for all the
        profile that are identical up to a permutation of the candidates.

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau.standardized_version
            TauVector({'a': Fraction(3, 10), 'b': Fraction(1, 10), 'bc': Fraction(3, 5)})
        """
        def translate(s, permute):
            return ''.join(sorted(s.replace('a', permute[0]).replace('b', permute[1]).replace('c', permute[2])))

        best_d = {}
        best_signature = []
        for perm in XYZ_PERMUTATIONS:
            d_test = {translate(ballot, perm): share for ballot, share in self.d_ballot_share.items()}
            signature_test = [d_test[ballot] for ballot in XYZ_BALLOTS_WITHOUT_INVERSION]
            if signature_test > best_signature:
                best_signature = signature_test
                best_d = d_test
        return TauVector({ballot: best_d[xyz_ballot]
                          for ballot, xyz_ballot in zip(BALLOTS_WITHOUT_INVERSIONS, XYZ_BALLOTS_WITHOUT_INVERSION)},
                         voting_rule=self.voting_rule)

    @cached_property
    def is_standardized(self):
        """Whether the profile is standardized or not. Cf. :meth:`standardized_version`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau.is_standardized
            False
        """
        return self == self.standardized_version

    # Pivots

    @cached_property
    def trio(self):
        """Event: trio."""
        return EventTrio(candidate_x='a', candidate_y='b', candidate_z='c',
                         tau_a=self.a, tau_b=self.b, tau_c=self.c, tau_ab=self.ab, tau_ac=self.ac, tau_bc=self.bc)

    @property
    def focus(self):
        """Focus : Focus of this tau-vector.

        This is based on the weak pivots.

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau.focus
            Focus.DIRECT
        """
        magnitudes = sorted([self.pivot_weak_ab.mu, self.pivot_weak_ac.mu, self.pivot_weak_bc.mu])
        if isclose(magnitudes[0], magnitudes[2]):
            return Focus.UNFOCUSED
        elif isclose(magnitudes[0], magnitudes[1]):
            return Focus.FORWARD_FOCUSED
        elif isclose(magnitudes[1], magnitudes[2]):
            return Focus.BACKWARD_FOCUSED
        else:
            return Focus.DIRECT

    def print_weak_pivots(self):
        """Print the weak pivots (including the 3-way tie).

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau.print_weak_pivots()
            pivot_weak_ab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
            pivot_weak_ac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
            pivot_weak_bc:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            trio:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
        """
        for pair in PAIRS_WITHOUT_INVERSIONS:
            print('pivot_weak_%s: ' % pair, getattr(self, 'pivot_weak_%s' % pair))
        print('trio: ', self.trio)

    def print_all_pivots(self):
        """Print all the pivots.

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau.print_all_pivots()
            pivot_weak_ab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
            pivot_weak_ac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
            pivot_weak_bc:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            pivot_strict_ab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
            pivot_strict_ac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
            pivot_strict_bc:  <asymptotic = exp(- inf)>
            pivot_tij_abc:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
            pivot_tij_acb:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.371758 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
            pivot_tij_bac:  <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
            pivot_tij_bca:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.302013 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            pivot_tij_cab:  <asymptotic = exp(- 0.0834849 n - 0.5 log n + 0.0518905 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
            pivot_tij_cba:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            pivot_tjk_abc:  <asymptotic = exp(- inf)>
            pivot_tjk_acb:  <asymptotic = exp(- inf)>
            pivot_tjk_bac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.371758 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
            pivot_tjk_bca:  <asymptotic = exp(- 0.0834849 n - 0.5 log n + 0.0518905 + o(1)), phi_a = 0.654654, \
phi_c = 1.52753, phi_ab = 0.654654>
            pivot_tjk_cab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
            pivot_tjk_cba:  <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
            trio:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            trio_1t_a:  <asymptotic = exp(- inf)>
            trio_1t_b:  <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.48597 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            trio_1t_c:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.490239 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            trio_2t_ab:  <asymptotic = exp(- 0.151472 n - 0.5 log n - 1.18339 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            trio_2t_ac:  <asymptotic = exp(- inf)>
            trio_2t_bc:  <asymptotic = exp(- 0.151472 n + 0.5 log n - 3.1394 + o(1)), phi_a = 0, phi_c = 1.41421, \
phi_ab = 0.707107>
            duo_ab:  <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
            duo_ac:  <asymptotic = exp(- 0.0834849 n - 0.5 log n - 0.87535 + o(1)), phi_a = 0.654654, phi_c = 1.52753, \
phi_ab = 0.654654>
            duo_bc:  <asymptotic = exp(- 0.0514719 n - 0.5 log n - 0.836813 + o(1)), phi_a = 1, phi_c = 1.41421, \
phi_ab = 0.707107>
        """
        for pair in PAIRS_WITHOUT_INVERSIONS:
            print('pivot_weak_%s: ' % pair, getattr(self, 'pivot_weak_%s' % pair))
        for pair in PAIRS_WITHOUT_INVERSIONS:
            print('pivot_strict_%s: ' % pair, getattr(self, 'pivot_strict_%s' % pair))
        for ranking in RANKINGS:
            print('pivot_tij_%s: ' % ranking, getattr(self, 'pivot_tij_%s' % ranking))
        for ranking in RANKINGS:
            print('pivot_tjk_%s: ' % ranking, getattr(self, 'pivot_tjk_%s' % ranking))
        print('trio: ', self.trio)
        for candidate in CANDIDATES:
            print('trio_1t_%s: ' % candidate, getattr(self, 'trio_1t_%s' % candidate))
        for pair in PAIRS_WITHOUT_INVERSIONS:
            print('trio_2t_%s: ' % pair, getattr(self, 'trio_2t_%s' % pair))
        for pair in PAIRS_WITHOUT_INVERSIONS:
            print('duo_%s: ' % pair, getattr(self, 'duo_%s' % pair))

    @cached_property
    def d_ranking_best_response(self):
        """dict : Best response profile.

        * Key: a ranking (e.g. ``'abc'``).
        * Value: a :class:`BestResponse` (whose subclass depends on :attr:`voting_rule`).

        Examples
        --------
            >>> from fractions import Fraction
            >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
            >>> tau.d_ranking_best_response['abc']
            <ballot = a, threshold_utility = 1, justification = Asymptotic method>
        """
        if self.voting_rule == APPROVAL:
            return DictPrintingInOrder({
                ranking: BestResponseApproval(tau=self, ranking=ranking) for ranking in RANKINGS})
        if self.voting_rule == PLURALITY:
            return DictPrintingInOrder({
                ranking: BestResponsePlurality(tau=self, ranking=ranking) for ranking in RANKINGS})
        if self.voting_rule == ANTI_PLURALITY:
            return DictPrintingInOrder({
                ranking: BestResponseAntiPlurality(tau=self, ranking=ranking) for ranking in RANKINGS})
        raise NotImplementedError

    @cached_property
    def score_ab_in_duo_ab(self):
        """Number : Common score of `a` and `b` in duo `ab`."""
        return (_multiply(self.a, self.duo_ab.phi_a)
                + _multiply(self.ab, self.duo_ab.phi_ab)
                + _multiply(self.ac, self.duo_ab.phi_ac))

    @cached_property
    def score_ac_in_duo_ac(self):
        """Number : Common score of `a` and `c` in duo `ac`."""
        return (_multiply(self.a, self.duo_ac.phi_a)
                + _multiply(self.ab, self.duo_ac.phi_ab)
                + _multiply(self.ac, self.duo_ac.phi_ac))

    @cached_property
    def score_bc_in_duo_bc(self):
        """Number : Common score of `b` and `c` in duo `bc`."""
        return (_multiply(self.b, self.duo_bc.phi_b)
                + _multiply(self.ab, self.duo_bc.phi_ab)
                + _multiply(self.bc, self.duo_bc.phi_bc))

    @cached_property
    def score_ba_in_duo_ba(self):
        """Number : Alternate notation for :attr:`score_ab_in_duo_ab`."""
        return self.score_ab_in_duo_ab

    @cached_property
    def score_ca_in_duo_ca(self):
        """Number : Alternate notation for :attr:`score_ac_in_duo_ac`."""
        return self.score_ac_in_duo_ac

    @cached_property
    def score_cb_in_duo_cb(self):
        """Number : Alternate notation for :attr:`score_bc_in_duo_bc`."""
        return self.score_bc_in_duo_bc

    @cached_property
    def score_c_in_duo_ab(self):
        """Number : Score of `c` in duo `ab`."""
        return (_multiply(self.c, self.duo_ab.phi_c)
                + _multiply(self.ac, self.duo_ab.phi_ac)
                + _multiply(self.bc, self.duo_ab.phi_bc))

    @cached_property
    def score_b_in_duo_ac(self):
        """Number : Score of `b` in duo `ac`."""
        return (_multiply(self.b, self.duo_ac.phi_b)
                + _multiply(self.ab, self.duo_ac.phi_ab)
                + _multiply(self.bc, self.duo_ac.phi_bc))

    @cached_property
    def score_a_in_duo_bc(self):
        """Number : Score of `a` in duo `bc`."""
        return (_multiply(self.a, self.duo_bc.phi_a)
                + _multiply(self.ab, self.duo_bc.phi_ab)
                + _multiply(self.ac, self.duo_bc.phi_ac))

    @cached_property
    def score_c_in_duo_ba(self):
        """Number : Alternate notation for :attr:`score_c_in_duo_ab`."""
        return self.score_c_in_duo_ab

    @cached_property
    def score_b_in_duo_ca(self):
        """Number : Alternate notation for :attr:`score_b_in_duo_ac`."""
        return self.score_b_in_duo_ac

    @cached_property
    def score_a_in_duo_cb(self):
        """Number : Alternate notation for :attr:`score_a_in_duo_bc`."""
        return self.score_a_in_duo_bc

    @cached_property
    def pivot_ab_easy_or_tight(self):
        """bool : True if the pivot `ab` is easy or tight, False if it is difficult."""
        pivot_easy = self.score_ab_in_duo_ab > self.score_c_in_duo_ab
        pivot_tight = isclose(self.score_ab_in_duo_ab, self.score_c_in_duo_ab)
        return pivot_easy or pivot_tight

    @cached_property
    def pivot_ac_easy_or_tight(self):
        """bool : True if the pivot `ac` is easy or tight, False if it is difficult."""
        pivot_easy = self.score_ac_in_duo_ac > self.score_b_in_duo_ac
        pivot_tight = isclose(self.score_ac_in_duo_ac, self.score_b_in_duo_ac)
        return pivot_easy or pivot_tight

    @cached_property
    def pivot_bc_easy_or_tight(self):
        """bool : True if the pivot `bc` is easy or tight, False if it is difficult."""
        pivot_easy = self.score_bc_in_duo_bc > self.score_a_in_duo_bc
        pivot_tight = isclose(self.score_bc_in_duo_bc, self.score_a_in_duo_bc)
        return pivot_easy or pivot_tight

    @cached_property
    def pivot_ba_easy_or_tight(self):
        """bool : Alternate notation for :attr:`pivot_ab_easy_or_tight`"""
        return self.pivot_ab_easy_or_tight

    @cached_property
    def pivot_ca_easy_or_tight(self):
        """bool : Alternate notation for :attr:`pivot_ac_easy_or_tight`"""
        return self.pivot_ac_easy_or_tight

    @cached_property
    def pivot_cb_easy_or_tight(self):
        """bool : Alternate notation for :attr:`pivot_bc_easy_or_tight`"""
        return self.pivot_bc_easy_or_tight


def _f_ballot_share(self, ballot):
    """Share of this ballot"""
    # This function is used to define an attribute for each ballot.
    return self.d_ballot_share[sort_ballot(ballot)]


for my_ballot in BALLOTS_WITH_INVERSIONS:
    setattr(TauVector, my_ballot, property(partial(_f_ballot_share, ballot=my_ballot)))
    if sort_ballot(my_ballot) == my_ballot:
        getattr(TauVector, my_ballot).__doc__ = "Number: Share of the ballot ``'%s'``." % my_ballot
    else:
        getattr(TauVector, my_ballot).__doc__ = \
            "Number: Share of the ballot ``'%s'`` (alternate notation)." % sort_ballot(my_ballot)

# Events based on a duo: create cached properties like duo_ab, etc.


def _f_duo(self, candidate_x, candidate_y, candidate_z, cls, stub):
    if candidate_x < candidate_y:
        return cls(candidate_x=candidate_x, candidate_y=candidate_y, candidate_z=candidate_z,
                   tau_a=self.a, tau_b=self.b, tau_c=self.c, tau_ab=self.ab, tau_ac=self.ac, tau_bc=self.bc)
    else:
        return getattr(self, stub + '_%s%s' % (candidate_y, candidate_x))


for event_class, event_stub, event_doc in [
        (EventDuo, 'duo', 'EventDuo : These two candidates have the same score.'),
        (EventPivotWeak, 'pivot_weak',
            'EventPivotWeak : These two candidates have the same score, at least as high as the other.'),
        (EventPivotStrict, 'pivot_strict',
            'EventPivotStrict : These two candidates have the same score, strictly higher than the other.'),
        (EventTrio2t, 'trio_2t',
            'EventTrio1t : These candidates have one vote less than the other.')]:
    for x, y, z in RANKINGS:
        name = event_stub + '_%s%s' % (x, y)
        setattr(TauVector, name, partial(_f_duo, candidate_x=x, candidate_y=y, candidate_z=z,
                                         cls=event_class, stub=event_stub))
        getattr(TauVector, name).__name__ = name
        setattr(TauVector, name, cached_property(getattr(TauVector, name)))
        getattr(TauVector, name).__doc__ = event_doc


# Events based on a permutation: create cached properties like pivot_tij_abc, etc.


def _f_ranking(self, candidate_x, candidate_y, candidate_z, cls):
    return cls(candidate_x=candidate_x, candidate_y=candidate_y, candidate_z=candidate_z,
               tau_a=self.a, tau_b=self.b, tau_c=self.c, tau_ab=self.ab, tau_ac=self.ac, tau_bc=self.bc)


for event_class, event_stub, event_doc in [
        (EventPivotTij, 'pivot_tij',
            'EventPivotTij: Personalized pivot of type Tij (between the two most-liked candidates).'),
        (EventPivotTjk, 'pivot_tjk',
            'EventPivotTjk: Personalized pivot of type Tjk (between the two least-liked candidates).')]:
    for x, y, z in RANKINGS:
        name = event_stub + '_%s%s%s' % (x, y, z)
        if event_stub == 'pivot_tjk':
            setattr(TauVector, name, partial(_f_ranking, candidate_x=z, candidate_y=y, candidate_z=x, cls=event_class))
        else:
            setattr(TauVector, name, partial(_f_ranking, candidate_x=x, candidate_y=y, candidate_z=z, cls=event_class))
        getattr(TauVector, name).__name__ = name
        setattr(TauVector, name, cached_property(getattr(TauVector, name)))
        getattr(TauVector, name).__doc__ = event_doc


# Events based on one candidate: create cached properties like trio_1t_a, etc.


for event_class, event_stub, event_doc in [
        (EventTrio1t, 'trio_1t', 'EventTrio1t : This candidate has one vote less than the two others.')]:
    for x, y, z in RANKINGS:
        if y > z:
            continue
        name = event_stub + '_%s' % x
        setattr(TauVector, name, partial(_f_ranking, candidate_x=x, candidate_y=y, candidate_z=z, cls=event_class))
        getattr(TauVector, name).__name__ = name
        setattr(TauVector, name, cached_property(getattr(TauVector, name)))
        getattr(TauVector, name).__doc__ = event_doc


# Used to compute easy and difficult pivots

def _multiply(tau_something, phi_something):
    """Return 0 if tau_something is 0, even if phi_something is nan."""
    return 0 if tau_something == 0 else tau_something * phi_something



from poisson_approval import ProfileOrdinal, PLURALITY, ANTI_PLURALITY


def test_normalization():
    profile = ProfileOrdinal(d_ranking_share={'abc': 1, 'acb': 1})
    assert profile.abc == 0.5


def test_plurality():
    """
        >>> profile = ProfileOrdinal(d_ranking_share={'abc': 1, 'acb': 1}, voting_rule=PLURALITY)
        >>> profile
        ProfileOrdinal({'abc': 0.5, 'acb': 0.5}, voting_rule='Plurality')
        >>> print(profile)
        <abc: 0.5, acb: 0.5> (Condorcet winner: a) (Plurality)
        >>> profile.tau_fanatic
        TauVector({'a': 1.0}, voting_rule='Plurality')
    """
    pass


def test_anti_plurality():
    """
        >>> profile = ProfileOrdinal(d_ranking_share={'abc': 1, 'acb': 1}, voting_rule=ANTI_PLURALITY)
        >>> profile
        ProfileOrdinal({'abc': 0.5, 'acb': 0.5}, voting_rule='Anti-plurality')
        >>> print(profile)
        <abc: 0.5, acb: 0.5> (Condorcet winner: a) (Anti-plurality)
        >>> profile.tau_fanatic
        TauVector({'ab': 0.5, 'ac': 0.5}, voting_rule='Anti-plurality')
    """
    pass
